import numba as nb
import numpy as np
from numba import njit

from src.Base.Poker.docs.index import *


@njit()
def getActionSize():
    return 6


@njit()
def getAgentSize():
    return 9


@njit()
def getStateSize():
    return 515


@njit()
def initEnv():
    env_state = np.zeros(ENV_LENGTH)
    env_state[ENV_CARD_OPEN:ENV_ALL_PLAYER_CHIP] = -1
    env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE] = np.full(
        NUMBER_PLAYER, 100 * BIG_CHIP
    )
    env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD] = np.ones(NUMBER_PLAYER)
    env_state[ENV_ALL_FIRST_CARD:ENV_BUTTON_PLAYER] = np.full(4 * NUMBER_PLAYER, -1)
    env_state[ENV_BUTTON_PLAYER] = -1  # người chơi ở vị trí đầu tiên giữ button
    return env_state


@njit()
def reset_round(old_env_state):
    env_state = np.zeros(ENV_LENGTH)
    #  print(np.sum(old_env_state[ENV_ALL_PLAYER_CHIP : ENV_ALL_PLAYER_CHIP_GIVE]), 'tổng chip')

    if (
        np.sum(
            old_env_state[ENV_ALL_PLAYER_CHIP : ENV_ALL_PLAYER_CHIP + NUMBER_PLAYER] > 0
        )
        == 1
    ):
        env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE] = old_env_state[
            ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE
        ]
        env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD] = (
            env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE] > 0
        ) * 1
        env_state[ENV_NUMBER_GAME_PLAYED] = old_env_state[ENV_NUMBER_GAME_PLAYED]
        #  print('ENDDDDDDDD GAME')
        #  print(np.sum(old_env_state[ENV_ALL_PLAYER_CHIP : ENV_ALL_PLAYER_CHIP_GIVE]), 'tổng chip')
        return env_state

    # tính toán chip còn lại của người chơi, từ đó tính ra trạng thái của người chơi ở game mới
    env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE] = old_env_state[
        ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE
    ]
    env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD] = (
        env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE] > 0
    ) * 1

    # thiết lập button dealer và temp button, trừ chip của small và big để cộng vào sum_pot_value, thiết lập id_action, status_game, phase, số ván đã chơi
    env_state[ENV_BUTTON_PLAYER] = (
        old_env_state[ENV_BUTTON_PLAYER] + 1
    ) % NUMBER_PLAYER
    while env_state[int(ENV_ALL_PLAYER_CHIP + env_state[ENV_BUTTON_PLAYER])] == 0:
        env_state[ENV_BUTTON_PLAYER] = (
            env_state[ENV_BUTTON_PLAYER] + 1
        ) % NUMBER_PLAYER
    sum_pot = 0
    # trừ chip của small_player
    sm_player = (env_state[ENV_BUTTON_PLAYER] + 1) % NUMBER_PLAYER
    while env_state[int(ENV_ALL_PLAYER_CHIP + sm_player)] == 0:
        sm_player = (sm_player + 1) % NUMBER_PLAYER
    sum_pot += SMALL_CHIP
    env_state[int(ENV_ALL_PLAYER_CHIP + sm_player)] -= SMALL_CHIP
    env_state[int(ENV_ALL_PLAYER_CHIP_GIVE + sm_player)] += SMALL_CHIP
    env_state[int(ENV_ALL_PLAYER_CHIP_IN_POT + sm_player)] += SMALL_CHIP

    # trừ chip của big_player
    big_player = (sm_player + 1) % NUMBER_PLAYER
    while env_state[int(ENV_ALL_PLAYER_CHIP + big_player)] == 0:
        big_player = (big_player + 1) % NUMBER_PLAYER
    sum_pot += min(BIG_CHIP, env_state[int(ENV_ALL_PLAYER_CHIP + big_player)])
    env_state[int(ENV_ALL_PLAYER_CHIP + big_player)] -= min(
        2, env_state[int(ENV_ALL_PLAYER_CHIP + big_player)]
    )
    env_state[int(ENV_ALL_PLAYER_CHIP_GIVE + big_player)] += min(
        2, env_state[int(ENV_ALL_PLAYER_CHIP + big_player)]
    )
    env_state[int(ENV_ALL_PLAYER_CHIP_IN_POT + big_player)] += min(
        2, env_state[int(ENV_ALL_PLAYER_CHIP + big_player)]
    )

    temp_button = (big_player + 1) % NUMBER_PLAYER
    while env_state[int(ENV_ALL_PLAYER_CHIP + temp_button)] == 0:
        temp_button = (temp_button + 1) % NUMBER_PLAYER
    env_state[ENV_TEMP_BUTTON] = temp_button

    env_state[ENV_CASH_TO_CALL_NEW] = BIG_CHIP
    env_state[ENV_CASH_TO_CALL_OLD] = BIG_CHIP
    env_state[ENV_POT_VALUE] = sum_pot

    # người chơi action kế tiếp là người thứ 1 trở đi từ button dealer mà còn chip
    env_state[ENV_ID_ACTION] = temp_button

    while env_state[int(ENV_ALL_PLAYER_CHIP + env_state[ENV_ID_ACTION])] == 0:
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    env_state[ENV_NUMBER_GAME_PLAYED] = old_env_state[ENV_NUMBER_GAME_PLAYED] + 1
    #  print('ván chơi thứ ', env_state[ENV_NUMBER_GAME_PLAYED])
    # reset 2 lá bài của người chơi và bài showdown, sau đó chia bài
    env_state[ENV_ALL_FIRST_CARD:ENV_BUTTON_PLAYER] = np.full(4 * NUMBER_PLAYER, -1)
    all_card_num = np.arange(52)
    np.random.shuffle(all_card_num)
    env_state[ENV_CARD_OPEN:ENV_ALL_PLAYER_CHIP] = all_card_num[:NUMBER_CARD_OPEN]
    env_state[ENV_ALL_FIRST_CARD:ENV_ALL_FIRST_CARD_SHOWDOWN] = all_card_num[
        NUMBER_CARD_OPEN : NUMBER_CARD_OPEN + 2 * NUMBER_PLAYER
    ]
    env_state[ENV_ALL_CARD_ON_BOARD:ENV_CARD_OPEN] = np.append(
        all_card_num[NUMBER_CARD_OPEN + 2 * NUMBER_PLAYER :],
        np.full(NUMBER_CARD_OPEN + 2 * NUMBER_PLAYER, -1),
    )
    return env_state


@njit()
def getAgentState(env_state):
    player_state = np.zeros(PLAYER_STATE_LENGTH)
    id_action = int(env_state[ENV_ID_ACTION])
    # cập nhật chip còn lại
    all_player_chip = env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE]
    player_state[P_ALL_PLAYER_CHIP:P_ALL_PLAYER_CHIP_GIVE] = np.concatenate(
        (all_player_chip[id_action:], all_player_chip[:id_action])
    )
    # cập nhật status
    all_player_status = env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
    player_state[P_ALL_PLAYER_STATUS:P_BUTTON_DEALER] = np.concatenate(
        (all_player_status[id_action:], all_player_status[:id_action])
    )
    # cập nhật game kết thúc chưa
    player_state[P_CHECK_END] = env_state[ENV_CHECK_END]
    player_state[P_NUMBER_GAME_PLAY] = env_state[ENV_NUMBER_GAME_PLAYED]

    if env_state[ENV_CHECK_END] == 0:
        # Cập nhật chip đã bỏ ra
        all_player_chip_give = env_state[
            ENV_ALL_PLAYER_CHIP_GIVE:ENV_ALL_PLAYER_CHIP_IN_POT
        ]
        player_state[P_ALL_PLAYER_CHIP_GIVE:P_ALL_PLAYER_STATUS] = np.concatenate(
            (all_player_chip_give[id_action:], all_player_chip_give[:id_action])
        )
        # cập nhật button dealer
        player_state[
            P_BUTTON_DEALER
            + int(env_state[ENV_BUTTON_PLAYER] - id_action) % NUMBER_PLAYER
        ] = 1
        # cập nhật chip to call
        player_state[P_CASH_TO_CALL] = max(
            0,
            env_state[ENV_CASH_TO_CALL_OLD]
            - env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action],
        )
        # cập nhật chip to bet
        player_state[P_CASH_TO_BET] = env_state[ENV_CASH_TO_CALL_OLD]
        # cập nhật pot value, phase, status game
        player_state[P_POT_VALUE] = env_state[ENV_POT_VALUE]
        player_state[P_PHASE] = env_state[ENV_PHASE]
        player_state[int(P_STATUS_GAME + max(env_state[ENV_STATUS_GAME] - 2, 0))] = 1

        # cập nhật card
        player_card = np.array(
            [
                env_state[ENV_ALL_FIRST_CARD + id_action],
                env_state[ENV_ALL_SECOND_CARD + id_action],
            ]
        ).astype(np.int64)
        if env_state[ENV_STATUS_GAME] == 0:
            player_state[P_ALL_CARD : P_ALL_CARD + NUMBER_CARD][player_card] = 1
        elif env_state[ENV_STATUS_GAME] != 6:
            # nếu ko phải showdown và pre flop
            card_open = env_state[
                ENV_CARD_OPEN : int(ENV_CARD_OPEN + env_state[ENV_STATUS_GAME])
            ].astype(np.int64)
            for id in range(1, NUMBER_PLAYER):
                status_id = player_state[P_ALL_PLAYER_STATUS:P_BUTTON_DEALER][id]
                if status_id == 1:
                    player_state[
                        P_ALL_CARD
                        + NUMBER_CARD * id : P_ALL_CARD
                        + NUMBER_CARD * (id + 1)
                    ][card_open] = 1

            player_card = np.append(player_card, card_open)
            player_state[P_ALL_CARD : P_ALL_CARD + NUMBER_CARD][player_card] = 1
        else:
            # duyệt từng người, ông nào tham gia showdown thì cho xem bài, ông nào bài xấu ko mở thì đã bị gán thành -1 ở step, nên lọc bài lớn hơn -1 là đc
            card_open = env_state[ENV_CARD_OPEN:ENV_ALL_PLAYER_CHIP].astype(np.int64)
            card_open = card_open[card_open > -1]  # thêm mới
            for id in range(NUMBER_PLAYER):
                status_id = player_state[P_ALL_PLAYER_STATUS:P_BUTTON_DEALER][id]
                if status_id == 1:
                    id_env = (id_action + id) % NUMBER_PLAYER
                    player_i_card = np.array(
                        [
                            env_state[ENV_ALL_FIRST_CARD_SHOWDOWN + id_env],
                            env_state[ENV_ALL_SECOND_CARD_SHOWDOWN + id_env],
                        ]
                    ).astype(np.int64)
                    player_i_card = np.append(player_i_card, card_open)
                    if id_env == id_action:
                        player_i_card = np.append(player_i_card, player_card)
                    player_i_card = player_i_card[player_i_card > -1]
                    player_state[
                        P_ALL_CARD
                        + NUMBER_CARD * id : P_ALL_CARD
                        + NUMBER_CARD * (id + 1)
                    ][player_i_card] = 1

    return player_state


@njit()
def getValidActions(player_state):
    list_action = np.zeros(6)
    if player_state[P_PHASE] == 0:
        # nếu cash to call == 0 thì check, bet, allin
        if player_state[P_CASH_TO_CALL] == 0:
            list_action[1:5] = 1
            list_action[2] = 0
        # nếu cash to call >= 0
        else:
            # nếu cash_to_call > cash (fold, allin)
            if player_state[P_ALL_PLAYER_CHIP] <= player_state[P_CASH_TO_CALL]:
                # fold, allin
                list_action[2] = 1
                list_action[4] = 1
            elif (
                player_state[P_ALL_PLAYER_CHIP] > player_state[P_CASH_TO_CALL]
                and player_state[P_ALL_PLAYER_CHIP] - player_state[P_CASH_TO_CALL]
                < player_state[P_CASH_TO_BET]
            ):
                # fold, call, allin
                list_action[0] = 1
                list_action[2] = 1
                list_action[4] = 1
            elif (
                player_state[P_ALL_PLAYER_CHIP] - player_state[P_CASH_TO_CALL]
                >= player_state[P_CASH_TO_BET]
            ):
                # call, fold, bet, allin
                list_action[:5] = 1
                list_action[1] = 0
                if (
                    np.count_nonzero(
                        player_state[P_ALL_PLAYER_CHIP:P_ALL_PLAYER_CHIP_GIVE]
                        * player_state[
                            P_ALL_PLAYER_STATUS : P_ALL_PLAYER_STATUS + NUMBER_PLAYER
                        ]
                    )
                    < 2
                ):
                    list_action[3:5] = 0

            else:
                raise Exception("xét thieu trường hợp")
    # nếu đang bet dở thì bet, allin, dừng bet
    elif player_state[P_PHASE] == 1:
        if player_state[P_ALL_PLAYER_CHIP] >= player_state[P_CASH_TO_BET]:
            list_action[3:] = 1
        else:
            list_action[4:] = 1
    return list_action


@njit()
def stepEnv(env_state, action):
    phase_env = int(env_state[ENV_PHASE])
    id_action = int(env_state[ENV_ID_ACTION])
    if phase_env == 0:
        if action == 0:  # người chơi call
            # trừ chip người chơi và tăng giá trị pot, bổ sung giá trị số tiền đã bỏ ra
            chip_to_call = (
                env_state[ENV_CASH_TO_CALL_OLD]
                - env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action]
            )
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] += chip_to_call
            env_state[ENV_ALL_PLAYER_CHIP_IN_POT + id_action] += chip_to_call
            env_state[ENV_ALL_PLAYER_CHIP + id_action] -= chip_to_call
            env_state[ENV_POT_VALUE] += chip_to_call
        elif action == 1:  # người chơi check
            pass
        elif action == 2:  # người chơi fold
            # nếu người chơi fold, chuyển trạng thái của người đó.
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] = 0
            env_state[ENV_ALL_PLAYER_STATUS + id_action] = 0
        elif action == 3:  # người chơi bet/raise
            # nếu người chơi bet, trừ tiền người chơi, tăng tiền pot, tăng cash to call_new
            # bổ sung giá trị số tiền đã bỏ ra
            chip_to_bet_raise = 0
            if env_state[ENV_CASH_TO_CALL_OLD] == 0:  # nếu chưa ai bet thì mình bet
                env_state[ENV_CASH_TO_CALL_OLD] = SMALL_CHIP
                chip_to_bet_raise = SMALL_CHIP
            else:  # nếu đã có người bet thì mình raise
                chip_to_bet_raise = (
                    2 * env_state[ENV_CASH_TO_CALL_OLD]
                    - env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action]
                )

            env_state[ENV_ALL_PLAYER_CHIP + id_action] -= chip_to_bet_raise
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] += chip_to_bet_raise
            env_state[ENV_ALL_PLAYER_CHIP_IN_POT + id_action] += chip_to_bet_raise
            env_state[ENV_POT_VALUE] += chip_to_bet_raise
            env_state[ENV_CASH_TO_CALL_NEW] += env_state[ENV_CASH_TO_CALL_OLD]
            # nếu bet thì sang phase xem có bet tiếp k

            env_state[ENV_TEMP_BUTTON] = id_action
            #  print('gán lại temp button', id_action, env_state[ENV_TEMP_BUTTON])
            env_state[ENV_PHASE] = 1
        elif action == 4:
            # nếu người chơi all_in, bổ sung giá trị chip người chơi bỏ ra và tăng pot
            chip_to_allin = env_state[ENV_ALL_PLAYER_CHIP + id_action]
            env_state[ENV_ALL_PLAYER_CHIP_IN_POT + id_action] += chip_to_allin
            env_state[ENV_POT_VALUE] += chip_to_allin
            #  print('check', chip_to_allin, env_state[ENV_CASH_TO_CALL_OLD])
            # điều chỉnh cash to call nếu cần, điều chỉnh button nếu người chơi bet nhiều chip nhất
            if env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] + chip_to_allin > np.max(
                env_state[ENV_ALL_PLAYER_CHIP_GIVE:ENV_ALL_PLAYER_CHIP_IN_POT]
            ):
                env_state[ENV_CASH_TO_CALL_OLD] = (
                    env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] + chip_to_allin
                )
                env_state[ENV_CASH_TO_CALL_NEW] = env_state[ENV_CASH_TO_CALL_OLD]
                env_state[ENV_TEMP_BUTTON] = id_action
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] += chip_to_allin
            # trừ hết tiền người chơi all in
            env_state[ENV_ALL_PLAYER_CHIP + id_action] = 0

        if action != 3:
            id_action_next = (id_action + 1) % NUMBER_PLAYER
            """
            hết lượt của người chơi hiện tại, chuyển người chơi, loop đến khi tìm đc người kế tiếp có thể action loop trở về đến
            vị trí hiện tại thì cũng dừng
            """
            if id_action_next != env_state[ENV_TEMP_BUTTON]:
                while (
                    env_state[ENV_ALL_PLAYER_STATUS + id_action_next] == 0
                    or env_state[ENV_ALL_PLAYER_CHIP + id_action_next] == 0
                ):
                    id_action_next = (id_action_next + 1) % NUMBER_PLAYER
                    #  print('while 1')
                    if id_action_next == env_state[ENV_TEMP_BUTTON]:
                        break
            #  print('check', id_action_next, env_state[ENV_STATUS_GAME], env_state[ENV_TEMP_BUTTON])
            if id_action_next == env_state[ENV_TEMP_BUTTON]:
                # kết thúc vòng chơi, chuyển status game, reset give chip, xác định id_action mới là người còn chơi và còn chip. Nếu hết 1 vòng => các người chơi đã allin hết, đi thẳng đến showdown
                if env_state[ENV_STATUS_GAME] != 5:
                    env_state[ENV_ALL_PLAYER_CHIP_GIVE:ENV_ALL_PLAYER_CHIP_IN_POT] = 0
                    env_state[ENV_ID_ACTION] = (
                        env_state[ENV_BUTTON_PLAYER] + 1
                    ) % NUMBER_PLAYER
                    env_state[ENV_CASH_TO_CALL_OLD], env_state[ENV_CASH_TO_CALL_NEW] = (
                        0,
                        0,
                    )

                    if (
                        np.count_nonzero(
                            env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
                        )
                        == 1
                        or np.count_nonzero(
                            env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE]
                            * env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
                        )
                        == 1
                    ):
                        if (
                            np.count_nonzero(
                                env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
                            )
                            == 1
                        ):
                            env_state[ENV_CARD_OPEN : ENV_CARD_OPEN + NUMBER_CARD_OPEN][
                                max(0, int(env_state[ENV_STATUS_GAME])) :
                            ] = -1
                        env_state[ENV_STATUS_GAME] = 6
                        #  print('gnas đi showdown khi chỉ còn 1 người còn tiền hoặc cả làng đã fold')
                        env_state[ENV_ID_ACTION] = env_state[ENV_TEMP_BUTTON]
                    else:
                        # loop cho đến khi gặp người vừa còn chơi và vừa còn chip
                        while (
                            env_state[
                                int(ENV_ALL_PLAYER_STATUS + env_state[ENV_ID_ACTION])
                            ]
                            == 0
                            or env_state[
                                int(ENV_ALL_PLAYER_CHIP + env_state[ENV_ID_ACTION])
                            ]
                            == 0
                        ):
                            #  print('while 2', env_state[ENV_ID_ACTION], env_state[ENV_BUTTON_PLAYER], env_state[int(ENV_ALL_PLAYER_STATUS + env_state[ENV_ID_ACTION])], env_state[int(ENV_ALL_PLAYER_CHIP + env_state[ENV_ID_ACTION])], env_state[ENV_BUTTON_PLAYER] )
                            env_state[ENV_ID_ACTION] = (
                                env_state[ENV_ID_ACTION] + 1
                            ) % NUMBER_PLAYER
                            # nếu mọi người hết khả năng action tiếp thì đi thẳng đến showdown
                            if env_state[ENV_ID_ACTION] == env_state[ENV_BUTTON_PLAYER]:
                                #  if np.count_nonzero(env_state[ENV_ALL_PLAYER_STATUS : ENV_ALL_FIRST_CARD]) != 1:
                                env_state[ENV_STATUS_GAME] = 6
                                #  print('gnas đi showdown')
                                env_state[ENV_ID_ACTION] = env_state[ENV_TEMP_BUTTON]
                                break

                    if env_state[ENV_STATUS_GAME] != 6:
                        temp_button = (env_state[ENV_BUTTON_PLAYER] + 1) % NUMBER_PLAYER
                        while env_state[int(ENV_ALL_PLAYER_STATUS + temp_button)] == 0:
                            temp_button = (temp_button + 1) % NUMBER_PLAYER
                        env_state[ENV_TEMP_BUTTON] = temp_button

                if env_state[ENV_STATUS_GAME] == 0:
                    env_state[ENV_STATUS_GAME] = 3
                elif env_state[ENV_STATUS_GAME] == 3:
                    env_state[ENV_STATUS_GAME] = 4
                elif env_state[ENV_STATUS_GAME] == 4:
                    env_state[ENV_STATUS_GAME] = 5
                elif env_state[ENV_STATUS_GAME] == 5:
                    env_state[ENV_STATUS_GAME] = 6

                if env_state[ENV_STATUS_GAME] == 6:
                    env_state = showdown(env_state)
                    env_state[ENV_ID_ACTION] = env_state[ENV_TEMP_BUTTON]
            else:
                # cập nhật người chơi
                env_state[ENV_ID_ACTION] = id_action_next
                # new thêm mới
                if (
                    np.count_nonzero(
                        env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
                    )
                    == 1
                ):
                    if env_state[ENV_STATUS_GAME] != 5:
                        env_state[ENV_CARD_OPEN : ENV_CARD_OPEN + NUMBER_CARD_OPEN][
                            max(0, int(env_state[ENV_STATUS_GAME])) :
                        ] = -1

                    env_state[ENV_STATUS_GAME] = 6
                    env_state = showdown(env_state)

    elif phase_env == 1:
        if action == 3:
            # nếu người chơi bet tiếp, trừ tiền người chơi, tăng tiền pot, tăng cash to call_new
            # bổ sung giá trị số tiền đã bỏ ra
            chip_to_bet_raise = env_state[ENV_CASH_TO_CALL_OLD]
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] += chip_to_bet_raise
            env_state[ENV_ALL_PLAYER_CHIP_IN_POT + id_action] += chip_to_bet_raise
            env_state[ENV_POT_VALUE] += chip_to_bet_raise
            env_state[ENV_CASH_TO_CALL_NEW] += chip_to_bet_raise
            env_state[ENV_ALL_PLAYER_CHIP + id_action] -= chip_to_bet_raise
        elif action == 4:
            # nếu all in khi bet tiếp, update chip give, pot value, cash_to_call_old and cash_to_call_new
            chip_to_allin = env_state[ENV_ALL_PLAYER_CHIP + id_action]
            env_state[ENV_ALL_PLAYER_CHIP + id_action] = 0
            env_state[ENV_ALL_PLAYER_CHIP_GIVE + id_action] += chip_to_allin
            env_state[ENV_ALL_PLAYER_CHIP_IN_POT + id_action] += chip_to_allin
            env_state[ENV_POT_VALUE] += chip_to_allin
            env_state[ENV_CASH_TO_CALL_OLD] = (
                env_state[ENV_CASH_TO_CALL_NEW] + chip_to_allin
            )
            env_state[ENV_CASH_TO_CALL_NEW] = env_state[ENV_CASH_TO_CALL_OLD]
            env_state[ENV_TEMP_BUTTON] = id_action
        elif action == 5:
            env_state[ENV_CASH_TO_CALL_OLD] = env_state[ENV_CASH_TO_CALL_NEW]
            env_state[ENV_CASH_TO_CALL_NEW] = env_state[ENV_CASH_TO_CALL_OLD]
            env_state[ENV_TEMP_BUTTON] = id_action

        if action != 3:
            """
            hết lượt của người chơi hiện tại, chuyển người chơi, loop đến khi tìm đc người kế tiếp có thể action loop trở về đến
            vị trí hiện tại thì cũng dừng
            """
            id_action_next = (id_action + 1) % NUMBER_PLAYER
            while (
                env_state[ENV_ALL_PLAYER_STATUS + id_action_next] == 0
                or env_state[ENV_ALL_PLAYER_CHIP + id_action_next] == 0
            ):
                id_action_next = (id_action_next + 1) % NUMBER_PLAYER
                #  print('while 3')
                if id_action_next == env_state[ENV_TEMP_BUTTON]:
                    break
            if id_action_next == env_state[ENV_TEMP_BUTTON]:
                # kết thúc vòng chơi, chuyển status game, reset give chip, xác định id_action mới là người còn chơi và còn chip. Nếu hết 1 vòng => các người chơi đã allin hết, đi thẳng đến showdown
                if env_state[ENV_STATUS_GAME] != 5:
                    env_state[ENV_ALL_PLAYER_CHIP_GIVE:ENV_ALL_PLAYER_CHIP_IN_POT] = 0
                    env_state[ENV_ID_ACTION] = (
                        env_state[ENV_BUTTON_PLAYER] + 1
                    ) % NUMBER_PLAYER
                    env_state[ENV_CASH_TO_CALL_OLD], env_state[ENV_CASH_TO_CALL_NEW] = (
                        0,
                        0,
                    )
                    # loop cho đến khi gặp người vừa còn chơi và vừa còn chip
                    while (
                        env_state[int(ENV_ALL_PLAYER_STATUS + env_state[ENV_ID_ACTION])]
                        == 0
                        or env_state[
                            int(ENV_ALL_PLAYER_CHIP + env_state[ENV_ID_ACTION])
                        ]
                        == 0
                    ):
                        env_state[ENV_ID_ACTION] = (
                            env_state[ENV_ID_ACTION] + 1
                        ) % NUMBER_PLAYER
                        # nếu mọi người hết khả năng action tiếp thì đi thẳng đến showdown
                        if env_state[ENV_ID_ACTION] == env_state[ENV_BUTTON_PLAYER]:
                            #  if np.count_nonzero(env_state[ENV_ALL_PLAYER_STATUS : ENV_ALL_FIRST_CARD]) != 1:
                            env_state[ENV_STATUS_GAME] = 6
                            #  print('gnas đi showdown')
                            env_state[ENV_ID_ACTION] = env_state[ENV_TEMP_BUTTON]
                            break
                    if env_state[ENV_STATUS_GAME] != 6:
                        temp_button = (env_state[ENV_BUTTON_PLAYER] + 1) % NUMBER_PLAYER
                        while env_state[int(ENV_ALL_PLAYER_STATUS + temp_button)] == 0:
                            temp_button = (temp_button + 1) % NUMBER_PLAYER
                        env_state[ENV_TEMP_BUTTON] = temp_button

                if env_state[ENV_STATUS_GAME] == 0:
                    env_state[ENV_STATUS_GAME] = 3
                elif env_state[ENV_STATUS_GAME] == 3:
                    env_state[ENV_STATUS_GAME] = 4
                elif env_state[ENV_STATUS_GAME] == 4:
                    env_state[ENV_STATUS_GAME] = 5
                elif env_state[ENV_STATUS_GAME] == 5:
                    env_state[ENV_STATUS_GAME] = 6

                if env_state[ENV_STATUS_GAME] == 6:
                    env_state = showdown(env_state)
                    env_state[ENV_ID_ACTION] = env_state[ENV_TEMP_BUTTON]

            else:
                # cập nhật người chơi
                env_state[ENV_ID_ACTION] = id_action_next
                env_state[ENV_PHASE] = 0

    return env_state


@njit()
def checkEnded(env_state):
    all_player_chip = env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE]
    all_player_status = env_state[ENV_ALL_PLAYER_STATUS:ENV_ALL_FIRST_CARD]
    number_game = env_state[ENV_NUMBER_GAME_PLAYED]
    if (
        np.count_nonzero(all_player_status) == 1
        and np.count_nonzero(all_player_chip) == 1
    ) or number_game == 100:
        return True
    return False


@njit()
def getReward(agent_state):
    all_player_chip = agent_state[P_ALL_PLAYER_CHIP:P_ALL_PLAYER_CHIP_GIVE]
    all_player_status = agent_state[P_ALL_PLAYER_STATUS:P_BUTTON_DEALER]

    if agent_state[P_CHECK_END] == 1 and (
        (
            np.count_nonzero(all_player_status) == 1
            and np.count_nonzero(all_player_chip) == 1
        )
        or agent_state[P_NUMBER_GAME_PLAY] == 100
    ):
        if np.argmax(all_player_chip) == 0:
            return 1
        else:
            return 0
    else:
        return -1


@njit()
def check_winner(env_state):
    all_player_chip = env_state[ENV_ALL_PLAYER_CHIP:ENV_ALL_PLAYER_CHIP_GIVE]
    winner = np.argmax(all_player_chip)
    return winner


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData

    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    action = arr_action[idx]
    if 0 in arr_action:
        action = 0
    elif 1 in arr_action:
        action = 1
    #  print(list(state))
    return action, perData


def one_game_print_mode(
    p0,
    list_other,
    per_player,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    env_state = initEnv()
    env_state = reset_round(env_state)
    while not checkEnded(env_state):
        print(
            "bài chung: ",
            ALL_CARD_STR[env_state[ENV_CARD_OPEN:ENV_ALL_PLAYER_CHIP].astype(np.int64)],
        )
        print(
            "bài riêng1: ",
            ALL_CARD_STR[
                env_state[ENV_ALL_FIRST_CARD:ENV_ALL_SECOND_CARD].astype(np.int64)
            ],
        )
        print(
            "bài riêng2: ",
            ALL_CARD_STR[
                env_state[ENV_ALL_SECOND_CARD:ENV_ALL_FIRST_CARD_SHOWDOWN].astype(
                    np.int64
                )
            ],
        )

        while env_state[ENV_STATUS_GAME] != 6:
            idx = int(env_state[ENV_ID_ACTION])
            player_state = getAgentState(env_state)
            if list_other[idx] == -1:
                action, per_player = p0(player_state, per_player)
                if getValidActions(player_state)[action] != 1:
                    raise Exception("bot dua ra action khong hop le")
            elif list_other[idx] == 1:
                action, per1 = p1(player_state, per1)
            elif list_other[idx] == 2:
                action, per2 = p2(player_state, per2)
            elif list_other[idx] == 3:
                action, per3 = p3(player_state, per3)
            elif list_other[idx] == 4:
                action, per4 = p4(player_state, per4)
            elif list_other[idx] == 5:
                action, per5 = p5(player_state, per5)
            elif list_other[idx] == 6:
                action, per6 = p6(player_state, per6)
            elif list_other[idx] == 7:
                action, per7 = p7(player_state, per7)
            elif list_other[idx] == 8:
                action, per8 = p8(player_state, per8)
            else:
                raise Exception("Sai list_other.")

            env_state = stepEnv(env_state, action)
        print(
            "turn bonus++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
        )
        for id in range(NUMBER_PLAYER):
            id_player = int(id + env_state[ENV_TEMP_BUTTON]) % NUMBER_PLAYER
            if env_state[ENV_ALL_PLAYER_STATUS + id_player] == 0:
                continue
            else:
                env_state[ENV_ID_ACTION] = id_player
                idx = int(env_state[ENV_ID_ACTION])
                player_state = getAgentState(env_state)
                if list_other[idx] == -1:
                    action, per_player = p0(player_state, per_player)
                    if getValidActions(player_state)[action] != 1:
                        raise Exception("bot dua ra action khong hop le")
                elif list_other[idx] == 1:
                    action, per1 = p1(player_state, per1)
                elif list_other[idx] == 2:
                    action, per2 = p2(player_state, per2)
                elif list_other[idx] == 3:
                    action, per3 = p3(player_state, per3)
                elif list_other[idx] == 4:
                    action, per4 = p4(player_state, per4)
                elif list_other[idx] == 5:
                    action, per5 = p5(player_state, per5)
                elif list_other[idx] == 6:
                    action, per6 = p6(player_state, per6)
                elif list_other[idx] == 7:
                    action, per7 = p7(player_state, per7)
                elif list_other[idx] == 8:
                    action, per8 = p8(player_state, per8)
                else:
                    raise Exception("Sai list_other.")
        print(
            "end turn bonus+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
        )
        env_state = reset_round(env_state)

    winner_real = check_winner(env_state)
    print("winner: ", winner_real, env_state[ENV_NUMBER_GAME_PLAYED])
    for p_idx in range(NUMBER_PLAYER):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        elif list_other[idx] == 5:
            action, per5 = p5(player_state, per5)
        elif list_other[idx] == 6:
            action, per6 = p6(player_state, per6)
        elif list_other[idx] == 7:
            action, per7 = p7(player_state, per7)
        elif list_other[idx] == 8:
            action, per8 = p8(player_state, per8)
        else:
            raise Exception("Sai list_other.")

        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4

    winner = False
    if np.where(list_other == -1)[0] == check_winner(env_state):
        winner = True
    else:
        winner = False
    return winner, per_player


def n_games_print_mode(
    p0,
    num_game,
    per_player,
    list_other,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_print_mode(
            p0,
            list_other,
            per_player,
            per1,
            per2,
            per3,
            per4,
            per5,
            per6,
            per7,
            per8,
            p1,
            p2,
            p3,
            p4,
            p5,
            p6,
            p7,
            p8,
        )
        win += winner
    return win, per_player


@njit()
def one_game_numba(
    p0,
    list_other,
    per_player,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    env_state = initEnv()
    env_state = reset_round(env_state)
    while not checkEnded(env_state):
        while env_state[ENV_STATUS_GAME] != 6:
            idx = int(env_state[ENV_ID_ACTION])
            player_state = getAgentState(env_state)
            if list_other[idx] == -1:
                action, per_player = p0(player_state, per_player)
                if getValidActions(player_state)[action] != 1:
                    raise Exception("bot dua ra action khong hop le")
            elif list_other[idx] == 1:
                action, per1 = p1(player_state, per1)
            elif list_other[idx] == 2:
                action, per2 = p2(player_state, per2)
            elif list_other[idx] == 3:
                action, per3 = p3(player_state, per3)
            elif list_other[idx] == 4:
                action, per4 = p4(player_state, per4)
            elif list_other[idx] == 5:
                action, per5 = p5(player_state, per5)
            elif list_other[idx] == 6:
                action, per6 = p6(player_state, per6)
            elif list_other[idx] == 7:
                action, per7 = p7(player_state, per7)
            elif list_other[idx] == 8:
                action, per8 = p8(player_state, per8)
            else:
                raise Exception("Sai list_other.")

            env_state = stepEnv(env_state, action)

        for id in range(NUMBER_PLAYER):
            id_player = int(id + env_state[ENV_TEMP_BUTTON]) % NUMBER_PLAYER
            if env_state[ENV_ALL_PLAYER_STATUS + id_player] == 0:
                continue
            else:
                env_state[ENV_ID_ACTION] = id_player
                idx = int(env_state[ENV_ID_ACTION])
                player_state = getAgentState(env_state)
                if list_other[idx] == -1:
                    action, per_player = p0(player_state, per_player)
                    if getValidActions(player_state)[action] != 1:
                        raise Exception("bot dua ra action khong hop le")
                elif list_other[idx] == 1:
                    action, per1 = p1(player_state, per1)
                elif list_other[idx] == 2:
                    action, per2 = p2(player_state, per2)
                elif list_other[idx] == 3:
                    action, per3 = p3(player_state, per3)
                elif list_other[idx] == 4:
                    action, per4 = p4(player_state, per4)
                elif list_other[idx] == 5:
                    action, per5 = p5(player_state, per5)
                elif list_other[idx] == 6:
                    action, per6 = p6(player_state, per6)
                elif list_other[idx] == 7:
                    action, per7 = p7(player_state, per7)
                elif list_other[idx] == 8:
                    action, per8 = p8(player_state, per8)
                else:
                    raise Exception("Sai list_other.")

        env_state = reset_round(env_state)

    env_state[ENV_CHECK_END] = 1
    win = check_winner(env_state)

    for p_idx in range(NUMBER_PLAYER):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        elif list_other[idx] == 5:
            action, per5 = p5(player_state, per5)
        elif list_other[idx] == 6:
            action, per6 = p6(player_state, per6)
        elif list_other[idx] == 7:
            action, per7 = p7(player_state, per7)
        elif list_other[idx] == 8:
            action, per8 = p8(player_state, per8)
        else:
            raise Exception("Sai list_other.")

        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    winner = False
    if np.where(list_other == -1)[0] == win:
        winner = True
    else:
        winner = False
    return winner, per_player


@njit()
def n_games_numba(
    p0,
    num_game,
    per_player,
    list_other,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0,
            list_other,
            per_player,
            per1,
            per2,
            per3,
            per4,
            per5,
            per6,
            per7,
            per8,
            p1,
            p2,
            p3,
            p4,
            p5,
            p6,
            p7,
            p8,
        )
        win += winner
    return win, per_player


def one_game_normal(
    p0,
    list_other,
    per_player,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    env_state = initEnv()
    env_state = reset_round(env_state)
    while not checkEnded(env_state):
        while env_state[ENV_STATUS_GAME] != 6:
            idx = int(env_state[ENV_ID_ACTION])
            player_state = getAgentState(env_state)
            if list_other[idx] == -1:
                action, per_player = p0(player_state, per_player)
                if getValidActions(player_state)[action] != 1:
                    raise Exception("bot dua ra action khong hop le")
            elif list_other[idx] == 1:
                action, per1 = p1(player_state, per1)
            elif list_other[idx] == 2:
                action, per2 = p2(player_state, per2)
            elif list_other[idx] == 3:
                action, per3 = p3(player_state, per3)
            elif list_other[idx] == 4:
                action, per4 = p4(player_state, per4)
            elif list_other[idx] == 5:
                action, per5 = p5(player_state, per5)
            elif list_other[idx] == 6:
                action, per6 = p6(player_state, per6)
            elif list_other[idx] == 7:
                action, per7 = p7(player_state, per7)
            elif list_other[idx] == 8:
                action, per8 = p8(player_state, per8)
            else:
                raise Exception("Sai list_other.")

            env_state = stepEnv(env_state, action)
        for id in range(NUMBER_PLAYER):
            id_player = int(id + env_state[ENV_TEMP_BUTTON]) % NUMBER_PLAYER
            if env_state[ENV_ALL_PLAYER_STATUS + id_player] == 0:
                continue
            else:
                env_state[ENV_ID_ACTION] = id_player
                idx = int(env_state[ENV_ID_ACTION])
                player_state = getAgentState(env_state)
                if list_other[idx] == -1:
                    action, per_player = p0(player_state, per_player)
                    if getValidActions(player_state)[action] != 1:
                        raise Exception("bot dua ra action khong hop le")
                elif list_other[idx] == 1:
                    action, per1 = p1(player_state, per1)
                elif list_other[idx] == 2:
                    action, per2 = p2(player_state, per2)
                elif list_other[idx] == 3:
                    action, per3 = p3(player_state, per3)
                elif list_other[idx] == 4:
                    action, per4 = p4(player_state, per4)
                elif list_other[idx] == 5:
                    action, per5 = p5(player_state, per5)
                elif list_other[idx] == 6:
                    action, per6 = p6(player_state, per6)
                elif list_other[idx] == 7:
                    action, per7 = p7(player_state, per7)
                elif list_other[idx] == 8:
                    action, per8 = p8(player_state, per8)
                else:
                    raise Exception("Sai list_other.")
        env_state = reset_round(env_state)

    env_state[ENV_CHECK_END] = 1
    win = check_winner(env_state)

    for p_idx in range(NUMBER_PLAYER):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            if getReward(player_state) == -1:
                print("game sai")
                print(list(env_state))
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        elif list_other[idx] == 5:
            action, per5 = p5(player_state, per5)
        elif list_other[idx] == 6:
            action, per6 = p6(player_state, per6)
        elif list_other[idx] == 7:
            action, per7 = p7(player_state, per7)
        elif list_other[idx] == 8:
            action, per8 = p8(player_state, per8)
        else:
            raise Exception("Sai list_other.")

        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    winner = False
    if np.where(list_other == -1)[0] == win:
        winner = True
    else:
        winner = False
    return winner, per_player


def n_games_normal(
    p0,
    num_game,
    per_player,
    list_other,
    per1,
    per2,
    per3,
    per4,
    per5,
    per6,
    per7,
    per8,
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0,
            list_other,
            per_player,
            per1,
            per2,
            per3,
            per4,
            per5,
            per6,
            per7,
            per8,
            p1,
            p2,
            p3,
            p4,
            p5,
            p6,
            p7,
            p8,
        )
        win += winner
    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def check_run_under_njit(agent, perData):
    return True


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_per_level_[4],
            _list_per_level_[5],
            _list_per_level_[6],
            _list_per_level_[7],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
            _list_bot_level_[4],
            _list_bot_level_[5],
            _list_bot_level_[6],
            _list_bot_level_[7],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_per_level_[4],
            _list_per_level_[5],
            _list_per_level_[6],
            _list_per_level_[7],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
            _list_bot_level_[4],
            _list_bot_level_[5],
            _list_bot_level_[6],
            _list_bot_level_[7],
        )



import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont

from src.Base.MachiKoro import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/MachiKoro/images/"
ROBBER_ICON = (40, 40)
BG_SIZE = (1200, 800)
CARD_SIZE = (60, 100)


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "background.webp").resize(BG_SIZE)
        self.robber = Image.open(IMG_PATH + "robber.jpg").resize(ROBBER_ICON)

        card_values = np.arange(19)
        self.cards = []
        for value in card_values:
            self.cards.append(Image.open(IMG_PATH + f"{value}.jpg").resize(CARD_SIZE))
        for value in card_values[-4:]:
            self.cards.append(
                Image.open(IMG_PATH + f"{value}down.jpg").resize(CARD_SIZE)
            )

        self.text_phase = [
            "chọn xúc sắc để đổ",
            "chọn đổ lại hay k",
            "chọn lấy tiền của ai",
            "chọn người để đổi",
            "chọn lá bài để đổi",
            "chọn lá bài muốn lấy",
            "chọn mua thẻ",
        ]


sprites = Sprites()


class Params:
    def __init__(self) -> None:
        self.center_card_x = BG_SIZE[0] * 0.5
        self.center_card_y = (BG_SIZE[1] - CARD_SIZE[1]) * 0.5
        self.list_coords_0 = [
            (self.center_card_x, 0.92 * BG_SIZE[1] - CARD_SIZE[1]),
            (0.82 * BG_SIZE[0], self.center_card_y),
            (self.center_card_x, 0.08 * BG_SIZE[1]),
            (0.18 * BG_SIZE[0], self.center_card_y),
        ]

        x_0 = BG_SIZE[0] * 0.32
        x_1 = BG_SIZE[0] * 0.68
        y_0 = 0.2 * BG_SIZE[1] - 0.25 * CARD_SIZE[1]
        y_1 = 0.8 * BG_SIZE[1] - 0.75 * CARD_SIZE[1]
        self.list_coords_1 = [(x_0, y_1), (x_1, y_1), (x_1, y_0), (x_0, y_0)]
        self.myFont = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=20
        )


params = Params()


_d_ = CARD_SIZE[0]


def draw_cards(bg, cards, s, y, rotate=False):
    n = cards.shape[0]
    y = round(y)
    #  print(rotate)
    if rotate % 2:
        for i in range(n):
            card = sprites.cards[cards[i]].rotate((rotate - 4) * 90, expand=True)
            bg.paste(
                card,
                (s, round(y + _d_ * i)),
            )
    else:
        for i in range(n):
            bg.paste(
                sprites.cards[cards[i]],
                (round(s + _d_ * i), y),
            )


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""
    else:
        action_mean = _env.ACTIONS_MEAN[action]

    return f"player {action_mean}"


class Env_components:
    def __init__(self, env, winner, list_other) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other


def get_env_components():
    env = _env.initEnv()
    winner = -1
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)
    env_components = Env_components(env, winner, list_other)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        env_components.env = _env.stepEnv(env_components.env, action)

    my_idx = np.where(env_components.list_other == -1)[0][0]
    env_components.winner = -1

    while _env.system_check_end(env_components.env):
        p_idx = int(env_components.env[_env.ENV_ID_ACTION])
        state = _env.getAgentState(env_components.env)

        if env_components.list_other[p_idx] == -1:
            win = -1
            return win, state, env_components

        agent = list_agent[env_components.list_other[p_idx] - 1]
        data = list_data[env_components.list_other[p_idx] - 1]
        action, data = agent(state, data)

        env_components.env = _env.stepEnv(env_components.env, action)

    env_components.env[_env.ENV_CHECK_END] = 1
    env_components.winner = _env.check_winner(env_components.env)
    for p_idx in range(4):
        if p_idx != my_idx:
            env_components.env[_env.ENV_ID_ACTION] = p_idx
            state = _env.getAgentState(env_components.env)

            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)

    env_components.env[_env.ENV_ID_ACTION] = my_idx
    state = _env.getAgentState(env_components.env)
    if my_idx == env_components.winner:
        win = 1
    else:
        win = 0

    return win, state, env_components


def get_state_image(state=None):
    background = sprites.background.copy()
    if state is None:
        return background

    draw = ImageDraw.Draw(background)
    myFont = params.myFont

    arr_player_card_coordinate = np.array(
        [
            [
                [250, 670],
                [310, 670],
                [370, 670],
                [430, 670],
                [490, 670],
                [550, 670],
                [610, 670],
                [670, 670],
                [730, 670],
                [790, 670],
                [850, 670],
                [910, 670],
                [370, 570],
                [430, 570],
                [490, 570],
                [550, 570],
                [610, 570],
                [670, 570],
                [730, 570],
            ],
            [
                [1080, 20],
                [1080, 80],
                [1080, 140],
                [1080, 200],
                [1080, 260],
                [1080, 320],
                [1080, 380],
                [1080, 440],
                [1080, 500],
                [1080, 560],
                [1080, 620],
                [1080, 680],
                [980, 140],
                [980, 200],
                [980, 260],
                [980, 320],
                [980, 380],
                [980, 440],
                [980, 500],
            ],
            [
                [250, 30],
                [310, 30],
                [370, 30],
                [430, 30],
                [490, 30],
                [550, 30],
                [610, 30],
                [670, 30],
                [730, 30],
                [790, 30],
                [850, 30],
                [910, 30],
                [370, 130],
                [430, 130],
                [490, 130],
                [550, 130],
                [610, 130],
                [670, 130],
                [730, 130],
            ],
            [
                [30, 20],
                [30, 80],
                [30, 140],
                [30, 200],
                [30, 260],
                [30, 320],
                [30, 380],
                [30, 440],
                [30, 500],
                [30, 560],
                [30, 620],
                [30, 680],
                [130, 140],
                [130, 200],
                [130, 260],
                [130, 320],
                [130, 380],
                [130, 440],
                [130, 500],
            ],
        ]
    )

    arr_count_card_coor = [
        [
            [270, 770],
            [330, 770],
            [390, 770],
            [450, 770],
            [510, 770],
            [570, 770],
            [630, 770],
            [690, 770],
            [750, 770],
            [810, 770],
            [870, 770],
            [930, 770],
        ],
        [
            [1180, 40],
            [1180, 100],
            [1180, 160],
            [1180, 220],
            [1180, 280],
            [1180, 340],
            [1180, 400],
            [1180, 460],
            [1180, 520],
            [1180, 580],
            [1180, 640],
            [1180, 700],
        ],
        [
            [270, 20],
            [330, 20],
            [390, 20],
            [450, 20],
            [510, 20],
            [570, 20],
            [630, 20],
            [690, 20],
            [750, 20],
            [810, 20],
            [870, 20],
            [930, 20],
        ],
        [
            [15, 40],
            [15, 100],
            [15, 160],
            [15, 220],
            [15, 280],
            [15, 340],
            [15, 400],
            [15, 460],
            [15, 520],
            [15, 580],
            [15, 640],
            [15, 700],
        ],
    ]
    for id in range(4):
        player_in4 = state[20 * id : 20 * (id + 1)]
        background = draw_card_player(
            player_in4,
            background,
            id,
            arr_player_card_coordinate,
            arr_count_card_coor,
            draw,
        )

    background = draw_card_on_board(state, background)

    background = draw_other(state, draw, background, myFont)

    return background


def draw_card_on_board(state, background):
    board_card = np.array(
        [
            [[370, 300], [430, 300], [490, 300], [550, 300], [610, 300], [670, 300]],
            [[370, 400], [430, 400], [490, 400], [550, 400], [610, 400], [670, 400]],
        ]
    )
    card_board = state[80:92] > 0
    for i in range(12):
        if card_board[i]:
            draw_cards(
                background,
                np.array([i]),
                board_card[i // 6][i % 6][0],
                board_card[i // 6][i % 6][1],
                rotate=0,
            )

    return background


def draw_card_player(
    player_in4,
    background,
    player_id,
    arr_player_card_coordinate,
    arr_count_card_coor,
    draw,
):
    card_player = player_in4[1:]
    # draw normal_card
    for i in range(19):
        if i < 15:
            if card_player[i] > 0:
                draw_cards(
                    background,
                    np.array([i]),
                    arr_player_card_coordinate[player_id][i][0],
                    arr_player_card_coordinate[player_id][i][1],
                    rotate=player_id,
                )
                if i < 12:
                    draw.text(arr_count_card_coor[player_id][i], f"{card_player[i]}")

        else:
            if card_player[i] > 0:
                draw_cards(
                    background,
                    np.array([i]),
                    arr_player_card_coordinate[player_id][i][0],
                    arr_player_card_coordinate[player_id][i][1],
                    rotate=player_id,
                )
            else:
                draw_cards(
                    background,
                    np.array([i + 4]),
                    arr_player_card_coordinate[player_id][i][0],
                    arr_player_card_coordinate[player_id][i][1],
                    rotate=player_id,
                )
    return background


def draw_other(state, draw, background, myFont):
    phase = np.where(state[122:129])[0][0] + 1
    last_dice_val = state[117]
    robber_coor = np.array([[550, 550], [940, 320], [550, 240], [240, 320]])
    coin_coor = np.array([[270, 640], [1000, 80], [270, 130], [130, 80]])
    player_coin = state[0:80:20]

    draw.text((750, 300), f"Phase:{phase}", (255, 0, 0), font=myFont)
    draw.text((750, 330), f"Add_turn:{state[116]}", (255, 0, 0), font=myFont)
    draw.text((750, 360), f"Val_dice:{last_dice_val}", (255, 0, 0), font=myFont)

    for id in range(4):
        draw.text(coin_coor[id], f"{int(player_coin[id])}đ", (255, 255, 0), font=myFont)

    if phase == 5:
        player_robbed = np.where(state[118:122])[0][0]
        background.paste(
            sprites.robber,
            (robber_coor[player_robbed][0], robber_coor[player_robbed][1]),
        )

    elif phase == 6:
        player_robbed = np.where(state[118:122])[0][0]
        background.paste(
            sprites.robber,
            (robber_coor[player_robbed][0], robber_coor[player_robbed][1]),
        )

        card_give = np.where(state[104:116])[0][0]
        card_give_coor = np.array(
            [
                [280, 720],
                [340, 720],
                [400, 720],
                [460, 720],
                [520, 720],
                [580, 720],
                [640, 720],
                [700, 720],
                [760, 720],
                [820, 720],
                [880, 720],
                [940, 720],
            ]
        )
        draw.text(card_give_coor[card_give], f"G", (255, 255, 255), font=myFont)

    elif phase == 7:
        card_buy_in_turn = state[92:104] > 0
        card_buy_coor = np.array(
            [
                [280, 720],
                [340, 720],
                [400, 720],
                [460, 720],
                [520, 720],
                [580, 720],
                [640, 720],
                [700, 720],
                [760, 720],
                [820, 720],
                [880, 720],
                [940, 720],
            ]
        )
        for i in range(len(card_buy_coor)):
            if card_buy_in_turn[i]:
                draw.text(card_buy_coor[i], f"V", (0, 255, 0), font=myFont)

    return background



import numpy as np
from PIL import Image, ImageDraw, ImageFont

from src.Base.Poker import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/Poker/images/"
BG_SIZE = (1680, 900)
CARD_SIZE = (70, 100)


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "background.png").resize(BG_SIZE)
        card_values = ["2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K", "A"]
        card_suits = ["Heart", "Diamond", "Club", "Spade"]
        self.cards = []
        for value in card_values:
            for suit in card_suits:
                self.cards.append(
                    Image.open(IMG_PATH + f"{value}-{suit}.png").resize(CARD_SIZE)
                )

        self.cards.append(Image.open(IMG_PATH + f"Card_back.png").resize(CARD_SIZE))

        #  self.other_in4 = []
        #  self.other_in4.append(Image.open(IMG_PATH+f"button_dealer.png").resize(CARD_SIZE))


sprites = Sprites()


class Params:
    def __init__(self) -> None:
        self.center_card_x = BG_SIZE[0] * 0.5
        self.center_card_y = (BG_SIZE[1] - CARD_SIZE[1]) * 0.5
        self.list_coords_0 = [
            (self.center_card_x, 0.92 * BG_SIZE[1] - CARD_SIZE[1]),
            (0.82 * BG_SIZE[0], self.center_card_y),
            (self.center_card_x, 0.08 * BG_SIZE[1]),
            (0.18 * BG_SIZE[0], self.center_card_y),
        ]

        x_0 = BG_SIZE[0] * 0.32
        x_1 = BG_SIZE[0] * 0.68
        y_0 = 0.2 * BG_SIZE[1] - 0.25 * CARD_SIZE[1]
        y_1 = 0.8 * BG_SIZE[1] - 0.75 * CARD_SIZE[1]
        self.list_coords_1 = [(x_0, y_1), (x_1, y_1), (x_1, y_0), (x_0, y_0)]
        self.myFont = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=40
        )
        self.fontFold = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=60
        )


params = Params()


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""
    else:
        action_mean = _env.ACTIONS_MEAN[action]

    return f"player {action_mean}"


class Env_components:
    def __init__(self, env, winner, list_other, showdown_yet) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other
        self.showdown_yet = True


def get_env_components():
    env = _env.initEnv()
    env = _env.reset_round(env)
    winner = -1
    showdown_yet = True
    list_other = np.array([-1, 1, 2, 3, 4, 5, 6, 7, 8])
    np.random.shuffle(list_other)
    env_components = Env_components(env, winner, list_other, showdown_yet)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        #  print('check qua 1')
        if env_components.env[_env.ENV_STATUS_GAME] != 6:
            env_components.env = _env.stepEnv(env_components.env, action)
    my_idx = np.where(env_components.list_other == -1)[0][0]
    #  print(f'bot là người chơi thứ {my_idx}')
    #  if _env.checkEnded(env_components.env):
    #      print('đoạn này thừa')
    #      env_components.env[_env.ENV_CHECK_END] = 1
    #      env_components.winner = _env.check_winner(env_components.env)

    while not _env.checkEnded(env_components.env):
        while env_components.env[_env.ENV_STATUS_GAME] != 6:
            p_idx = int(env_components.env[_env.ENV_ID_ACTION])

            if env_components.list_other[p_idx] == -1:
                #  print('toang đây')
                state = _env.getAgentState(env_components.env)
                win = -1
                return win, state, env_components

            state = _env.getAgentState(env_components.env)
            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)
            if _env.getValidActions(state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")
            #  print(f'người chơi {p_idx} action {action}')

            env_components.env = _env.stepEnv(env_components.env, action)

        # chạy turn showdown
        if env_components.showdown_yet:
            #  print('chạy show down')
            my_idx = -1
            for id in range(_env.NUMBER_PLAYER):
                env_components.env[_env.ENV_ID_ACTION] = id
                p_idx = int(env_components.env[_env.ENV_ID_ACTION])
                if env_components.env[_env.ENV_ALL_PLAYER_STATUS + p_idx] == 0:
                    continue
                else:
                    if env_components.list_other[p_idx] == -1:
                        my_idx = p_idx
                        continue
                    state = _env.getAgentState(env_components.env)
                    agent = list_agent[env_components.list_other[p_idx] - 1]
                    data = list_data[env_components.list_other[p_idx] - 1]
                    action, data = agent(state, data)
            env_components.showdown_yet = False
            if env_components.env[_env.ENV_ALL_PLAYER_STATUS + my_idx] == 1:
                #  print('check đây', my_idx)
                env_components.env[_env.ENV_ID_ACTION] = my_idx
                state = _env.getAgentState(env_components.env)
                win = -1
                return win, state, env_components
        else:
            #  print('reset round')
            env_components.env = _env.reset_round(env_components.env)
            env_components.showdown_yet = True

    env_components.env[_env.ENV_CHECK_END] = 1
    env_components.winner = _env.check_winner(env_components.env)

    #  for id in range(_env.NUMBER_PLAYER):
    #      env_components.env[_env.ENV_PHASE] = 1
    #      p_idx = int(env_components.env[_env.ENV_ID_ACTION])
    #      state = _env.getAgentState(env_components.env)
    #      my_idx = np.where(env_components.list_other == -1)[0][0]
    #      if my_idx == env_components.winner:
    #          win = 1
    #      else:
    #          win = 0

    my_idx = np.where(env_components.list_other == -1)[0][0]
    env_components.env[_env.ENV_PHASE] = 1
    # chạy turn bonus cho bot hệ thống
    for p_idx in range(_env.NUMBER_PLAYER):
        if p_idx == my_idx:
            continue
        else:
            env_components.env[_env.ENV_ID_ACTION] = p_idx
            _state = _env.getAgentState(env_components.env)
            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(_state, data)
    env_components.env[_env.ENV_ID_ACTION] = my_idx
    state = _env.getAgentState(env_components.env)
    if my_idx == env_components.winner:
        win = 1
    else:
        win = 0

    return win, state, env_components


_d_ = CARD_SIZE[0] * 0.2


def draw_cards(bg, cards, s, y, back=False, faded=False):
    n = cards.shape[0]
    y = round(y)
    if back:
        if faded:
            im = sprites.faded_card_back
        else:
            im = sprites.card_back

        for i in range(n):
            bg.paste(im, (round(s + _d_ * i), y))
    else:
        for i in range(n):
            bg.paste(sprites.cards[cards[i]], (round(s + _d_ * i), y))


def get_state_image(state=None):
    # mảng tọa độ in ảnh
    arr_coordinate_card = np.array(
        [
            [400, 720],
            [1000, 720],
            [1300, 550],
            [1300, 300],
            [1050, 100],
            [700, 100],
            [350, 100],
            [150, 300],
            [150, 550],
        ]
    )
    arr_coordinate_button_dealer = np.array(
        [
            [440, 680],
            [1040, 680],
            [1270, 570],
            [1270, 330],
            [1070, 200],
            [740, 200],
            [390, 200],
            [250, 330],
            [250, 570],
        ]
    )
    arr_coordinate_small = np.array(
        [
            [400, 680],
            [1000, 680],
            [1270, 540],
            [1270, 300],
            [1040, 200],
            [710, 200],
            [360, 200],
            [250, 300],
            [250, 540],
        ]
    )
    arr_coordinate_big = np.array(
        [
            [480, 680],
            [1080, 680],
            [1270, 600],
            [1270, 360],
            [1100, 200],
            [770, 200],
            [420, 200],
            [250, 360],
            [250, 600],
        ]
    )
    arr_coordinate_status = np.array(
        [
            [420, 750],
            [1020, 750],
            [1320, 560],
            [1320, 320],
            [1080, 120],
            [720, 120],
            [370, 120],
            [180, 320],
            [180, 560],
        ]
    )
    arr_coordinate_chip_else = np.array(
        [
            [420, 830],
            [1020, 830],
            [1380, 560],
            [1380, 330],
            [1080, 70],
            [720, 70],
            [370, 70],
            [80, 330],
            [80, 560],
        ]
    )
    arr_coordinate_chip_in_pot = np.array(
        [
            [320, 800],
            [920, 800],
            [1320, 520],
            [1320, 270],
            [980, 120],
            [620, 120],
            [270, 120],
            [180, 270],
            [180, 520],
        ]
    )

    background = sprites.background.copy()
    draw = ImageDraw.Draw(background)
    myFont = params.myFont
    fontFold = params.fontFold

    if state is None:
        return background
    # lấy thông tin (trạng thái người chơi, tiền còn, tiền đã bỏ ra trong turn)
    arr_player_status = state[486:495]
    arr_player_chip_in_pot = state[477:486]
    arr_player_chip_else = state[468:477]
    player_in = (arr_player_chip_in_pot + arr_player_chip_else + arr_player_status) > 0

    # Lấy card bàn chơi
    open_card = np.array([])
    check = 0
    for i in range(9):
        player_i_card = np.where(state[52 * i : 52 * (i + 1)])[0]
        if len(open_card) == 0:
            open_card = player_i_card.copy()
            check += 1
        elif len(player_i_card) > 0 and len(open_card) > 0:
            open_card = np.intersect1d(open_card, player_i_card)
            check += 1
            break
    if check < 2:
        open_card = np.array([])
    #  print(open_card,  'card chung', 'số ván đã chơi', state[513], state[508:513])
    # lấy card cho người chơi
    for i in range(9):
        player_i_card = np.where(state[52 * i : 52 * (i + 1)])[0]
        if len(player_i_card) < 1 or len(player_i_card) == len(open_card):
            if player_in[i]:
                player_i_card = np.array([-1, -1])
            else:
                player_i_card = np.array([])
        else:
            player_i_card = np.setdiff1d(player_i_card, open_card)
            if len(player_i_card) == 1:
                player_i_card = np.append(player_i_card, -1)

        draw_cards(
            background,
            player_i_card,
            arr_coordinate_card[i][0],
            arr_coordinate_card[i][1],
        )
        draw.text(
            (arr_coordinate_chip_else[i][0], arr_coordinate_chip_else[i][1]),
            f"{int(state[468+i])}",
            (255, 0, 0),
            font=myFont,
        )

    open_card_coodinate = np.array(
        [[640, 400], [720, 400], [800, 400], [880, 400], [960, 400]]
    )
    for i in range(len(open_card)):
        draw_cards(
            background,
            open_card[i : i + 1],
            open_card_coodinate[i][0],
            open_card_coodinate[i][1],
        )
    # DEALER, SMALL, BIG, STATUS
    if np.sum(player_in) > 1 and state[513] < 100:
        dealer = np.where(state[495:504])[0][0]
        draw.text(
            (
                arr_coordinate_button_dealer[dealer][0],
                arr_coordinate_button_dealer[dealer][1],
            ),
            f"D",
            (255, 255, 255),
            font=myFont,
        )
        check_BS = 0
        id_player = (dealer + 1) % 9
        while check_BS < 2:
            if player_in[id_player] == 1:
                if check_BS == 0:
                    draw.text(
                        (
                            arr_coordinate_small[id_player][0],
                            arr_coordinate_small[id_player][1],
                        ),
                        f"S",
                        (255, 0, 0),
                        font=myFont,
                    )
                else:
                    draw.text(
                        (
                            arr_coordinate_big[id_player][0],
                            arr_coordinate_big[id_player][1],
                        ),
                        f"B",
                        (255, 255, 0),
                        font=myFont,
                    )
                check_BS += 1
            id_player = (id_player + 1) % 9

        for i in range(9):
            if arr_player_status[i] == 0:
                draw.text(
                    (arr_coordinate_status[i][0], arr_coordinate_status[i][1]),
                    f"F",
                    (255, 0, 0),
                    font=fontFold,
                )
            else:
                draw.text(
                    (
                        arr_coordinate_chip_in_pot[i][0],
                        arr_coordinate_chip_in_pot[i][1],
                    ),
                    f"{int(state[477 + i])}",
                    (255, 165, 0),
                    font=myFont,
                )

    draw.text((700, 500), f"Pot:{int(state[506])}", (255, 0, 0), font=myFont)
    return background



import sys
import numpy as np
from numba import njit
from numba.typed import List

import env

game_name = sys.argv[1]
env.make(game_name)


def convert_to_save(perData):
    if len(perData) == 2:
        raise Exception("Data này đã được convert rồi.")

    data = List()
    data.append(np.zeros((1, env.getActionSize())))
    temp = np.zeros((env.getActionSize(), env.getActionSize()))
    for i in range(temp.shape[0]):
        temp[i] = np.argsort(np.argsort(perData[2][i])) + 1e-6 * np.random.rand(
            env.getActionSize()
        )

    data.append(temp)
    return data


def convert_to_test(perData):
    return List(perData)


@njit()
def DataAgent():
    return List(
        [
            np.zeros((1, env.getActionSize())),
            np.random.rand(env.getActionSize(), env.getActionSize()),
            np.zeros((env.getActionSize(), env.getActionSize())),
        ]
    )


@njit()
def Train(state, per):
    actions = env.getValidActions(state)
    weight = per[0][0]

    output = actions * weight + actions
    c = np.where(output == np.max(output))[0]
    action = c[np.random.randint(0, c.shape[0])]

    per[0] += per[1][action]
    win = env.getReward(state)

    if win != -1:
        per[0][:, :] = 0.0
        if win == 1:
            per[2] += per[1]
        else:
            per[1] = np.random.rand(env.getActionSize(), env.getActionSize())

    return action, per


@njit()
def Test(state, per):
    actions = env.getValidActions(state)
    weight = per[0][0]

    output = actions * weight + actions
    action = np.argmax(output)

    weight[:] += per[1][action]
    win = env.getReward(state)
    if win != -1:
        per[0][:, :] = 0.0
    return action, per



import importlib.util
import math

import ipywidgets as widgets
import numpy as np
from IPython.display import clear_output, display

from env import SHORT_PATH


def import_files(game_name):
    spec = importlib.util.spec_from_file_location(
        "_env", f"{SHORT_PATH}src/Base/{game_name}/env.py"
    )
    global _env
    _env = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(_env)

    spec_1 = importlib.util.spec_from_file_location(
        "_render_func", f"{SHORT_PATH}src/Base/{game_name}/_render_func.py"
    )
    global _render_func
    _render_func = importlib.util.module_from_spec(spec_1)
    spec_1.loader.exec_module(_render_func)


class Render:
    def __init__(self, Agent, Data, list_agent, list_data, max_temp_frame) -> None:
        self.list_agent = list_agent
        self.list_data = list_data
        self.Agent = Agent
        self.Data = Data
        self.max_temp_frame = max_temp_frame

        self.output_image = widgets.Output()

        self.take = widgets.Dropdown(
            options=["Take an action"],
            value="Take an action",
            disabled=True,
            layout=widgets.Layout(width="20%"),
        )
        self.take.observe(self.handle_take, "value")

        self.explain = widgets.Dropdown(
            options=["Explain an action"],
            value="Explain an action",
            disabled=True,
            layout=widgets.Layout(width="20%"),
        )
        self.explain.observe(self.handle_explain, "value")

        self.slider = widgets.IntSlider(
            value=-1,
            min=-1,
            max=0,
            step=1,
            description="State:",
            disabled=True,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            layout=widgets.Layout(width="50%"),
        )
        self.slider.observe(self.handle_slider, "value")

        self.previous = widgets.Button(
            disabled=True, icon="fa-chevron-left", layout=widgets.Layout(width="5%")
        )
        self.previous.on_click(self.handle_previous)

        self.next = widgets.Button(
            disabled=True, icon="fa-chevron-right", layout=widgets.Layout(width="5%")
        )
        self.next.on_click(self.handle_next)

        self.output_text = widgets.Output()

        self.hbox = widgets.HBox(
            [self.take, self.previous, self.slider, self.next, self.explain],
            layout=widgets.Layout(width="100%"),
        )
        self.vbox = widgets.VBox(
            [self.output_image, self.hbox, self.output_text],
            layout=widgets.Layout(display="flex", align_items="center"),
        )

        self.start()

    def start(self):
        self.env_components = _render_func.get_env_components()
        self.history_state = []
        self.images = {}
        self.max_state_idx = -1
        self.history_action = []
        self.history_valid_actions = []

        if type(self.Agent) == str and self.Agent.lower() == "human":
            self.system_mode = "play"
            self.step()
        else:
            action = None
            win, state, self.env_components = _render_func.get_main_player_state(
                self.env_components, self.list_agent, self.list_data, action
            )
            self.history_state.append(state)
            while win == -1:
                valid_actions = np.where(_env.getValidActions(state) == 1)[0]
                self.history_valid_actions.append(valid_actions)
                action, self.Data = self.Agent(state, self.Data)
                if action not in valid_actions:
                    raise Exception("Invalid action")

                self.history_action.append(action)
                win, state, self.env_components = _render_func.get_main_player_state(
                    self.env_components, self.list_agent, self.list_data, action
                )
                self.history_state.append(state)

            if win == 0:
                self.system_mode = "You lose"
            else:
                self.system_mode = "You win"

            self.max_state_idx = len(self.history_state) - 1
            self.slider.max = self.max_state_idx
            self.slider.value = self.slider.max
            self.slider.min = 0

    def step(self, action=None):
        self.disable_all()

        if not action is None:
            self.history_action.append(action)

        win, state, self.env_components = _render_func.get_main_player_state(
            self.env_components, self.list_agent, self.list_data, action
        )
        self.history_state.append(state)
        self.max_state_idx += 1

        if win == 0:
            self.system_mode = "You lose"
        elif win == 1:
            self.system_mode = "You win"

        if len(self.images) == self.max_temp_frame:
            min_ = min(self.images.keys())
            self.images.pop(min_)

        self.images[self.max_state_idx] = _render_func.get_state_image(state)
        self.slider.max = self.max_state_idx
        self.slider.value = self.slider.max
        if self.max_state_idx == 0:
            self.slider.min = 0

    def render(self):
        return display(self.vbox)

    def handle_take(self, p):
        if p.new != "Take an action":
            self.step(p.new)

        self.take.value = "Take an action"

    def handle_explain(self, p):
        if p.new != "Explain an action":
            self.show_text(f"{p.new}: {_render_func.get_description(p.new)}")

        self.explain.value = "Explain an action"

    def handle_previous(self, p):
        self.slider.value -= 1

    def handle_next(self, p):
        self.slider.value += 1

    def handle_slider(self, p):
        self.disable_all()

        if p.new not in self.images.keys():
            if p.new >= self.max_state_idx / 2.0:
                up = int(p.new + self.max_temp_frame / 2.0)
                upbound = min(self.max_state_idx, up)
                l_ = self.max_temp_frame - (upbound - p.new + 1)
                lowbound = p.new - l_
                lowbound = max(0, lowbound)
            else:
                low = math.ceil(p.new - self.max_temp_frame / 2.0)
                lowbound = max(0, low)
                u_ = self.max_temp_frame - (p.new - lowbound + 1)
                upbound = p.new + u_
                upbound = min(self.max_state_idx, upbound)

            keys = list(self.images.keys())
            for key in keys:
                if key < lowbound or key > upbound:
                    self.images.pop(key)

            keys = list(self.images.keys())
            for key in range(lowbound, upbound + 1):
                if key not in keys:
                    self.images[key] = _render_func.get_state_image(
                        self.history_state[key]
                    )

        self.show_image(p.new)

        if self.system_mode == "play" and p.new == self.max_state_idx:
            state = self.history_state[p.new]
            if len(self.history_valid_actions) != self.max_state_idx + 1:
                valid_actions = np.where(_env.getValidActions(state) == 1)[0]
                self.history_valid_actions.append(valid_actions)
            else:
                valid_actions = self.history_valid_actions[p.new]

            self.take.options = ["Take an action"] + list(valid_actions)
            self.explain.options = ["Explain an action"] + list(valid_actions)
            self.show_text("")
            self.take.disabled = False
            self.explain.disabled = False
            if p.new != 0:
                self.previous.disabled = False
        else:
            if p.new != self.max_state_idx:
                valid_actions = self.history_valid_actions[p.new]
                self.explain.options = ["Explain an action"] + list(valid_actions)
                self.show_text(f"Action has been taken: {self.history_action[p.new]}")
                self.next.disabled = False
                self.explain.disabled = False
            else:
                self.show_text(self.system_mode)

            if p.new != 0:
                self.previous.disabled = False

        self.slider.disabled = False

    def show_image(self, idx):
        with self.output_image:
            clear_output(wait=True)
            display(self.images[idx])

    def show_text(self, txt):
        with self.output_text:
            clear_output(wait=True)
            print(txt)

    def disable_all(self):
        self.take.disabled = True
        self.explain.disabled = True
        self.slider.disabled = True
        self.previous.disabled = True
        self.next.disabled = True



import numpy as np
from numba import njit

# Splendor Game Environment
#  normal_cards_infor = np.array([[0, 2, 2, 2, 0, 0, 0], [0, 2, 3, 0, 0, 0, 0], [0, 2, 1, 1, 0, 2, 1], [0, 2, 0, 1, 0, 0, 2], [0, 2, 0, 3, 1, 0, 1], [0, 2, 1, 1, 0, 1, 1], [1, 2, 0, 0, 0, 4, 0], [0, 2, 2, 1, 0, 2, 0], [0, 1, 2, 0, 2, 0, 1], [0, 1, 0, 0, 2, 2, 0], [0, 1, 1, 0, 1, 1, 1], [0, 1, 2, 0, 1, 1, 1], [0, 1, 1, 1, 3, 0, 0], [0, 1, 0, 0, 0, 2, 1], [0, 1, 0, 0, 0, 3, 0], [1, 1, 4, 0, 0, 0, 0], [1, 0, 0, 0, 0, 0, 4], [0, 0, 0, 0, 0, 0, 3], [0, 0, 0, 1, 1, 1, 2], [0, 0, 0, 0, 1, 2, 2], [0, 0, 1, 0, 0, 3, 1], [0, 0, 2, 0, 0, 0, 2], [0, 0, 0, 1, 1, 1, 1], [0, 0, 0, 2, 1, 0, 0], [0, 4, 0, 2, 2, 1, 0], [0, 4, 1, 1, 2, 1, 0], [0, 4, 0, 1, 0, 1, 3], [1, 4, 0, 0, 4, 0, 0], [0, 4, 0, 2, 0, 2, 0], [0, 4, 2, 0, 0, 1, 0], [0, 4, 1, 1, 1, 1, 0], [0, 4, 0, 3, 0, 0, 0], [0, 3, 1, 0, 2, 0, 0], [0, 3, 1, 1, 1, 0, 1], [1, 3, 0, 4, 0, 0, 0], [0, 3, 1, 2, 0, 0, 2], [0, 3, 0, 0, 3, 0, 0], [0, 3, 0, 0, 2, 0, 2], [0, 3, 3, 0, 1, 1, 0], [0, 3, 1, 2, 1, 0, 1], [1, 2, 0, 3, 0, 2, 2], [2, 2, 0, 2, 0, 1, 4], [1, 2, 3, 0, 2, 0, 3], [2, 2, 0, 5, 3, 0, 0], [2, 2, 0, 0, 5, 0, 0], [3, 2, 0, 0, 6, 0, 0], [3, 1, 0, 6, 0, 0, 0], [2, 1, 1, 0, 0, 4, 2], [2, 1, 0, 5, 0, 0, 0], [2, 1, 0, 3, 0, 0, 5], [1, 1, 0, 2, 3, 3, 0], [1, 1, 3, 2, 2, 0, 0], [3, 0, 6, 0, 0, 0, 0], [2, 0, 0, 0, 0, 5, 3], [2, 0, 0, 0, 0, 5, 0], [2, 0, 0, 4, 2, 0, 1], [1, 0, 2, 3, 0, 3, 0], [1, 0, 2, 0, 0, 3, 2], [3, 4, 0, 0, 0, 0, 6], [2, 4, 5, 0, 0, 3, 0], [2, 4, 5, 0, 0, 0, 0], [1, 4, 3, 3, 0, 0, 2], [1, 4, 2, 0, 3, 2, 0], [2, 4, 4, 0, 1, 2, 0], [1, 3, 0, 2, 2, 0, 3], [1, 3, 0, 0, 3, 2, 3], [2, 3, 2, 1, 4, 0, 0], [2, 3, 3, 0, 5, 0, 0], [2, 3, 0, 0, 0, 0, 5], [3, 3, 0, 0, 0, 6, 0], [4, 2, 0, 7, 0, 0, 0], [4, 2, 0, 6, 3, 0, 3], [5, 2, 0, 7, 3, 0, 0], [3, 2, 3, 3, 0, 3, 5], [3, 1, 3, 0, 3, 5, 3], [4, 1, 0, 0, 0, 0, 7], [5, 1, 0, 3, 0, 0, 7], [4, 1, 0, 3, 0, 3, 6], [3, 0, 0, 5, 3, 3, 3], [4, 0, 0, 0, 7, 0, 0], [5, 0, 3, 0, 7, 0, 0], [4, 0, 3, 3, 6, 0, 0], [5, 4, 0, 0, 0, 7, 3], [3, 4, 5, 3, 3, 3, 0], [4, 4, 0, 0, 0, 7, 0], [4, 4, 3, 0, 0, 6, 3], [3, 3, 3, 3, 5, 0, 3], [5, 3, 7, 0, 0, 3, 0], [4, 3, 6, 0, 3, 3, 0], [4, 3, 7, 0, 0, 0, 0]])
normal_cards_infor = np.array(
    [
        [0, 0, 0, 1, 0, 0, 2, 2, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 3, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 1, 1, 0, 2, 1],
        [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 2],
        [0, 0, 0, 1, 0, 0, 0, 3, 1, 0, 1],
        [0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1],
        [1, 0, 0, 1, 0, 0, 0, 0, 0, 4, 0],
        [0, 0, 0, 1, 0, 0, 2, 1, 0, 2, 0],
        [0, 0, 1, 0, 0, 0, 2, 0, 2, 0, 1],
        [0, 0, 1, 0, 0, 0, 0, 0, 2, 2, 0],
        [0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1],
        [0, 0, 1, 0, 0, 0, 2, 0, 1, 1, 1],
        [0, 0, 1, 0, 0, 0, 1, 1, 3, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 2, 1],
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 3, 0],
        [1, 0, 1, 0, 0, 0, 4, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 4],
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 3],
        [0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 2],
        [0, 1, 0, 0, 0, 0, 0, 0, 1, 2, 2],
        [0, 1, 0, 0, 0, 0, 1, 0, 0, 3, 1],
        [0, 1, 0, 0, 0, 0, 2, 0, 0, 0, 2],
        [0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 1, 0, 0, 0, 0, 0, 2, 1, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 2, 2, 1, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 2, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 3],
        [1, 0, 0, 0, 0, 1, 0, 0, 4, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 2, 0, 2, 0],
        [0, 0, 0, 0, 0, 1, 2, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 3, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 0, 2, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1],
        [1, 0, 0, 0, 1, 0, 0, 4, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1, 2, 0, 0, 2],
        [0, 0, 0, 0, 1, 0, 0, 0, 3, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0, 2, 0, 2],
        [0, 0, 0, 0, 1, 0, 3, 0, 1, 1, 0],
        [0, 0, 0, 0, 1, 0, 1, 2, 1, 0, 1],
        [1, 0, 0, 1, 0, 0, 0, 3, 0, 2, 2],
        [2, 0, 0, 1, 0, 0, 0, 2, 0, 1, 4],
        [1, 0, 0, 1, 0, 0, 3, 0, 2, 0, 3],
        [2, 0, 0, 1, 0, 0, 0, 5, 3, 0, 0],
        [2, 0, 0, 1, 0, 0, 0, 0, 5, 0, 0],
        [3, 0, 0, 1, 0, 0, 0, 0, 6, 0, 0],
        [3, 0, 1, 0, 0, 0, 0, 6, 0, 0, 0],
        [2, 0, 1, 0, 0, 0, 1, 0, 0, 4, 2],
        [2, 0, 1, 0, 0, 0, 0, 5, 0, 0, 0],
        [2, 0, 1, 0, 0, 0, 0, 3, 0, 0, 5],
        [1, 0, 1, 0, 0, 0, 0, 2, 3, 3, 0],
        [1, 0, 1, 0, 0, 0, 3, 2, 2, 0, 0],
        [3, 1, 0, 0, 0, 0, 6, 0, 0, 0, 0],
        [2, 1, 0, 0, 0, 0, 0, 0, 0, 5, 3],
        [2, 1, 0, 0, 0, 0, 0, 0, 0, 5, 0],
        [2, 1, 0, 0, 0, 0, 0, 4, 2, 0, 1],
        [1, 1, 0, 0, 0, 0, 2, 3, 0, 3, 0],
        [1, 1, 0, 0, 0, 0, 2, 0, 0, 3, 2],
        [3, 0, 0, 0, 0, 1, 0, 0, 0, 0, 6],
        [2, 0, 0, 0, 0, 1, 5, 0, 0, 3, 0],
        [2, 0, 0, 0, 0, 1, 5, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 1, 3, 3, 0, 0, 2],
        [1, 0, 0, 0, 0, 1, 2, 0, 3, 2, 0],
        [2, 0, 0, 0, 0, 1, 4, 0, 1, 2, 0],
        [1, 0, 0, 0, 1, 0, 0, 2, 2, 0, 3],
        [1, 0, 0, 0, 1, 0, 0, 0, 3, 2, 3],
        [2, 0, 0, 0, 1, 0, 2, 1, 4, 0, 0],
        [2, 0, 0, 0, 1, 0, 3, 0, 5, 0, 0],
        [2, 0, 0, 0, 1, 0, 0, 0, 0, 0, 5],
        [3, 0, 0, 0, 1, 0, 0, 0, 0, 6, 0],
        [4, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0],
        [4, 0, 0, 1, 0, 0, 0, 6, 3, 0, 3],
        [5, 0, 0, 1, 0, 0, 0, 7, 3, 0, 0],
        [3, 0, 0, 1, 0, 0, 3, 3, 0, 3, 5],
        [3, 0, 1, 0, 0, 0, 3, 0, 3, 5, 3],
        [4, 0, 1, 0, 0, 0, 0, 0, 0, 0, 7],
        [5, 0, 1, 0, 0, 0, 0, 3, 0, 0, 7],
        [4, 0, 1, 0, 0, 0, 0, 3, 0, 3, 6],
        [3, 1, 0, 0, 0, 0, 0, 5, 3, 3, 3],
        [4, 1, 0, 0, 0, 0, 0, 0, 7, 0, 0],
        [5, 1, 0, 0, 0, 0, 3, 0, 7, 0, 0],
        [4, 1, 0, 0, 0, 0, 3, 3, 6, 0, 0],
        [5, 0, 0, 0, 0, 1, 0, 0, 0, 7, 3],
        [3, 0, 0, 0, 0, 1, 5, 3, 3, 3, 0],
        [4, 0, 0, 0, 0, 1, 0, 0, 0, 7, 0],
        [4, 0, 0, 0, 0, 1, 3, 0, 0, 6, 3],
        [3, 0, 0, 0, 1, 0, 3, 3, 5, 0, 3],
        [5, 0, 0, 0, 1, 0, 7, 0, 0, 3, 0],
        [4, 0, 0, 0, 1, 0, 6, 0, 3, 3, 0],
        [4, 0, 0, 0, 1, 0, 7, 0, 0, 0, 0],
    ]
)
noble_cards_infor = np.array(
    [
        [0, 4, 4, 0, 0],
        [3, 0, 3, 3, 0],
        [3, 3, 3, 0, 0],
        [3, 0, 0, 3, 3],
        [0, 3, 0, 3, 3],
        [4, 0, 4, 0, 0],
        [4, 0, 0, 4, 0],
        [0, 3, 3, 0, 3],
        [0, 4, 0, 0, 4],
        [0, 0, 0, 4, 4],
    ]
)

P_STATE_SIZE = 221
P_ACTION_SIZE = 15
P_SCORE = 17

P_END_BOARD_INFOR = 6
P_START_AGENT_INFOR = 6
P_END_AGENT_INFOR = 18
P_START_CARD_NORMAL = 18
P_END_CARD_NORMAL = 150
P_START_CARD_NOBLE = 150
P_END_CARD_NOBLE = 175
P_START_CARD_UPSIDE_DOWN = 175
P_END_CARD_UPSIDE_DOWN = 208
P_START_GET_RES = 208
P_END_GET_RES = 213
P_START_OTHER_SCORE = 213
P_END_OTHER_SCORE = 216
P_START_UPSIDE_DOWN_HIDE_CARD = 216
P_END_UPSIDE_DOWN_HIDE_CARD = 219
P_COUNT_CAN_UPSIDE_DOWN = 219
P_CLOSE_GAME = 220

P_START_STOCK = 6
P_END_STOCK = 11
P_YELLOW_STOCK = 11
P_START_STOCK_COUNT = 12
P_END_STOCK_COUNT = 17
TOTAL_INFOR_NORMAL_CARD = 11


@njit()
def initEnv():
    env_state = np.full(164, 0)
    env_state[:] = 0
    env_state[101:107] = np.array([7, 7, 7, 7, 7, 5])
    lv1 = np.arange(40)
    lv2 = np.arange(40, 70)
    lv3 = np.arange(70, 90)
    nob = np.arange(90, 100)
    for lv in [lv1, lv2, lv3]:
        np.random.shuffle(lv)
        env_state[lv[:4]] = 5
    np.random.shuffle(nob)
    env_state[nob[:5]] = 5
    env_state[161] = lv1[4]
    env_state[162] = lv2[4]
    env_state[163] = lv3[4]

    return env_state, lv1, lv2, lv3


@njit()
def get_list_id_card_on_lv(lv):
    if len(lv) >= 4:
        return lv[:4]
    else:
        return lv[: len(lv)]


@njit()
def checkEnded(env_state):
    score_arr = np.array([env_state[118 + 12 * p_id] for p_id in range(4)])
    max_score = np.max(score_arr)
    if max_score >= 15 and env_state[100] % 4 == 0:
        lst_p = np.where(score_arr == max_score)[0] + 1
        if len(lst_p) == 1:
            return lst_p[0]
        else:
            lst_p_c = []
            for p_id in lst_p:
                lst_p_c.append(np.count_nonzero(env_state[:90] == p_id))

            lst_p_c = np.array(lst_p_c)
            min_p_c = np.min(lst_p_c)
            lst_p_win = np.where(lst_p_c == min_p_c)[0]
            if len(lst_p_win) == 1:
                return lst_p[lst_p_win[0]]
            else:
                id_max = -1
                a = -1
                for i in lst_p_win:
                    b = max(np.where(env_state[:90] == lst_p[i])[0])
                    if b > a:
                        id_max = lst_p[i]
                        a = b

                return id_max

    else:
        return 0


@njit()
def concatenate_all_lv_card(lv1, lv2, lv3):
    card_lv1 = normal_cards_infor[get_list_id_card_on_lv(lv1)]
    card_lv2 = normal_cards_infor[get_list_id_card_on_lv(lv2)]
    card_lv3 = normal_cards_infor[get_list_id_card_on_lv(lv3)]
    list_open_card = np.append(card_lv1, card_lv2)
    list_open_card = np.append(list_open_card, card_lv3)
    return list_open_card


@njit()
def get_id_card_normal_in_lv(lv1, lv2, lv3):
    list_card_normal_on_board = np.append(
        get_list_id_card_on_lv(lv1), get_list_id_card_on_lv(lv2)
    )
    list_card_normal_on_board = np.append(
        list_card_normal_on_board, get_list_id_card_on_lv(lv3)
    )
    return list_card_normal_on_board


@njit()
def getAgentState(env_state, lv1, lv2, lv3):
    p_id = env_state[100] % 4  # Lấy người đang chơi
    b_infor = env_state[101:107]  #  Lấy 6 loại nguyên liệu của bàn chơi
    p_infor = env_state[
        107 + 12 * p_id : 119 + 12 * p_id
    ]  # Lấy thông tin người đang chơi, 6 nguyên liệu trên bàn, 5 nguyên liệu mặc định, điểm

    list_open_card = concatenate_all_lv_card(
        lv1, lv2, lv3
    )  # Lấy list thẻ normal đang mở trên bàn
    list_open_noble = noble_cards_infor[
        np.where(env_state[90:100] == 5)
    ].flatten()  # Lấy list thẻ Noble đang mở trên bàn

    state_card_normal = np.full(132, 0)
    state_card_noble = np.full(25, 0)
    state_card_normal[: len(list_open_card)] = list_open_card
    state_card_noble[: len(list_open_noble)] = list_open_noble

    list_upside_down_card = normal_cards_infor[np.where(env_state[:90] == -(p_id + 1))]
    p_upside_down_card = np.full(33, 0)
    if len(list_upside_down_card) > 0:
        array_hide_card = list_upside_down_card.flatten()
        p_upside_down_card[: len(array_hide_card)] = array_hide_card

    st_getting = env_state[155:160]  # Lấy thông tin 5 nguyên liệu đang lấy trong turn
    other_scores = [
        env_state[118 + 12 * id_other_player]
        for id_other_player in range(4)
        if id_other_player != p_id
    ]  # Lấy điểm của người chơi khác

    p_state = np.zeros(P_STATE_SIZE)
    p_state[:P_END_BOARD_INFOR] = b_infor
    p_state[P_START_AGENT_INFOR:P_END_AGENT_INFOR] = p_infor
    p_state[
        P_START_CARD_NORMAL:P_END_CARD_NORMAL
    ] = state_card_normal  # Lấy thông tin 12 thẻ đang mở ở trên bàn
    p_state[P_START_CARD_NOBLE:P_END_CARD_NOBLE] = state_card_noble
    p_state[
        P_START_CARD_UPSIDE_DOWN:P_END_CARD_UPSIDE_DOWN
    ] = p_upside_down_card  # Lấy thông tin 3 thẻ đang úp
    p_state[
        P_START_GET_RES:P_END_GET_RES
    ] = st_getting  # Lấy thông tin 5 nguyên liệu đang lấy trong turn
    p_state[
        P_START_OTHER_SCORE:P_END_OTHER_SCORE
    ] = other_scores  # Lấy điểm của người chơi khác
    p_state[P_START_UPSIDE_DOWN_HIDE_CARD:P_END_UPSIDE_DOWN_HIDE_CARD] = (
        env_state[161:164] != 100
    ) * 1  # Lấy thông tin của các thẻ ẩn có thẻ úp, nếu có thể úp thì là 1
    p_state[P_COUNT_CAN_UPSIDE_DOWN] = len(
        np.where(env_state[:90] == 5)[0]
    )  # Số lượng thẻ có thể úp trong bàn

    cls_game = int(checkEnded(env_state))
    if cls_game == 0:
        p_state[P_CLOSE_GAME] = 0
    else:
        p_state[P_CLOSE_GAME] = 1
    return p_state.astype(np.float64)


@njit()
def getReward(p_state):
    scores = p_state[P_START_OTHER_SCORE:P_END_OTHER_SCORE]
    owner_score = p_state[P_SCORE]

    if p_state[P_CLOSE_GAME] == 0:
        return -1
    if owner_score >= 15 and max(scores) <= owner_score:
        return 1
    if max(scores) >= 15 and max(scores) > owner_score:
        return 0
    if p_state[P_CLOSE_GAME] == 1:
        return 0


@njit()
def get_remove_card_on_lv_and_add_new_card(
    env_state, lv, p_id, id_card_hide, type_action, card_id
):
    if type_action == 2:
        env_state[lv[4]] = -(p_id + 1)
        id_card_in_level = 4
    else:
        if len(lv) > 4:
            env_state[lv[4]] = 5
        id_card_in_level = np.where(lv == card_id)[0][0]
        if type_action == 1:
            env_state[card_id] = p_id + 1

    lv = np.delete(lv, id_card_in_level)
    if len(lv) > 4:
        env_state[id_card_hide] = lv[4]
    else:
        env_state[id_card_hide] = 100
    return env_state, lv


@njit()
def get_card_can_get(
    env_state, p_id, cur_p, b_stocks, card_id, nl_auto, nl_bt, card_infor, lv1, lv2, lv3
):
    cur_p[:5] -= nl_bt
    cur_p[5] -= nl_auto
    b_stocks[:5] += nl_bt
    b_stocks[5] += nl_auto

    x_ = env_state[card_id]
    env_state[card_id] = p_id + 1
    if x_ == 5:  # Type_action == 1
        if card_id < 40:
            env_state, lv1 = get_remove_card_on_lv_and_add_new_card(
                env_state, lv1, p_id, 161, 1, card_id
            )
        elif card_id >= 40 and card_id < 70:
            env_state, lv2 = get_remove_card_on_lv_and_add_new_card(
                env_state, lv2, p_id, 162, 1, card_id
            )
            env_state[card_id] = p_id + 1
        else:
            env_state, lv3 = get_remove_card_on_lv_and_add_new_card(
                env_state, lv3, p_id, 163, 1, card_id
            )

    cur_p[6:11][np.where(card_infor[1:6] == 1)[0][0]] += 1  # const_stock
    cur_p[11] += card_infor[0]  # Score

    #  Check Noble
    noble_lst = []
    nobles = [i for i in range(90, 100) if env_state[:100][i] == 5]
    for noble_id in nobles:
        if (noble_cards_infor[noble_id - 90][-5:] <= cur_p[6:11]).all():
            noble_lst.append(noble_id)

    for noble_id in noble_lst:
        env_state[noble_id] = p_id + 1
        cur_p[11] += 3

    env_state[107 + 12 * p_id : 119 + 12 * p_id] = cur_p
    env_state[101:107] = b_stocks

    return env_state, lv1, lv2, lv3


@njit()
def get_card(env_state, action_, cur_p, b_stocks, p_id, lv1, lv2, lv3):
    if action_ < 12:
        id_action = int(action_)
        id_card_normal_ = get_id_card_normal_in_lv(lv1, lv2, lv3)
    else:
        #   print('Thẻ này đang úp')
        id_action = int(action_) - 12
        id_card_normal_ = np.where(env_state[:90] == -(p_id + 1))[0]

    card_id = id_card_normal_[id_action]
    card_infor = normal_cards_infor[card_id]
    card_price = card_infor[-5:]
    nl_bo_ra = (card_price > cur_p[6:11]) * (card_price - cur_p[6:11])
    nl_bt = np.minimum(nl_bo_ra, cur_p[:5])
    nl_auto = np.sum(nl_bo_ra - nl_bt)

    card_need = cur_p[:5] + cur_p[6:11] - card_price

    #   print('Taget', get_id_card(card_id))
    if (
        -sum(card_need[np.where(card_need < 0)]) <= cur_p[5] or min(card_need) >= 0
    ):  # (x*x>0)
        #   print('Lấy thẻ', card_infor, card_need, get_id_card(card_id))

        return get_card_can_get(
            env_state,
            p_id,
            cur_p,
            b_stocks,
            card_id,
            nl_auto,
            nl_bt,
            card_infor,
            lv1,
            lv2,
            lv3,
        )


@njit()
def return_res_to_board(env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id):
    cur_p[:5] += array_res_buy
    b_stocks[:5] -= array_res_buy
    env_state[107 + 12 * p_id : 119 + 12 * p_id] = cur_p
    env_state[101:107] = b_stocks
    return env_state, lv1, lv2, lv3


@njit()
def return_2_res_to_board(env_state, lv1, lv2, lv3, cur_p, b_stocks, res_, p_id):
    cur_p[:5][res_] += 2
    b_stocks[:5][res_] -= 2

    env_state[107 + 12 * p_id : 119 + 12 * p_id] = cur_p
    env_state[101:107] = b_stocks
    return env_state, lv1, lv2, lv3


@njit()
def stepEnv(action, env_state, lv1, lv2, lv3, all_actions):
    all_actions = np.where(all_actions == 1)[0]
    p_id = env_state[100] % 4
    cur_p = env_state[107 + 12 * p_id : 119 + 12 * p_id]
    b_stocks = env_state[101:107]
    env_state[100] += 1
    if action < 12:
        id_action = action
        id_card_normal = get_id_card_normal_in_lv(lv1, lv2, lv3)
    else:
        #   print('Thẻ này đang úp')
        id_action = action - 12
        id_card_normal = np.where(env_state[:90] == -(p_id + 1))[0]

    card_id = id_card_normal[id_action]
    card_infor = normal_cards_infor[card_id]
    card_price = card_infor[-5:]
    nl_bo_ra = (card_price > cur_p[6:11]) * (card_price - cur_p[6:11])
    nl_bt = np.minimum(nl_bo_ra, cur_p[:5])
    nl_auto = np.sum(nl_bo_ra - nl_bt)
    #   print(nl_bo_ra)

    card_need = cur_p[:5] + cur_p[6:11] - card_price

    #   print('Taget', get_id_card(card_id))
    if (
        -np.sum(card_need[np.where(card_need < 0)]) <= cur_p[5]
        or np.min(card_need) >= 0
    ):  # (x*x>0)
        #   print('Lấy thẻ', card_infor, card_need, cur_p[5], get_id_card(card_id))
        return get_card_can_get(
            env_state,
            p_id,
            cur_p,
            b_stocks,
            card_id,
            nl_auto,
            nl_bt,
            card_infor,
            lv1,
            lv2,
            lv3,
        )

    res_max = np.argmax(nl_bo_ra)
    if np.sum(cur_p[:6]) <= 8:
        if np.max(nl_bo_ra - cur_p[:5]) >= 2 and b_stocks[res_max] >= 4:
            #   print('Lấy 2 nguyên liệu', res_max, card_infor, nl_bo_ra)
            return return_2_res_to_board(
                env_state, lv1, lv2, lv3, cur_p, b_stocks, res_max, p_id
            )

        for res in np.argsort(nl_bo_ra)[::-1]:
            if (nl_bo_ra[res] - cur_p[:5][res]) >= 2 and b_stocks[res] >= 4:
                #   print('Lấy 2 nguyên liệu nhiều thứ 2__', res, card_infor, nl_bo_ra)
                return return_2_res_to_board(
                    env_state, lv1, lv2, lv3, cur_p, b_stocks, res, p_id
                )

    res_board_have = b_stocks[:5]
    res_can_buy = np.where(((nl_bo_ra - cur_p[:5]) > 0) & (res_board_have > 0), 1, 0)
    array_res_buy = np.full(5, 0)
    if np.sum(cur_p[:6]) <= 7:
        for id, res in enumerate(res_can_buy):
            if res == 1:
                array_res_buy[id] = 1
                if np.sum(array_res_buy) == 3:
                    #   print('Mua nguyên liệu_I:', array_res_buy, '----', res_can_buy, card_infor, nl_bo_ra)
                    return return_res_to_board(
                        env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
                    )

        res_du = np.where(((res_board_have > 0) - array_res_buy) > 0)[0]
        if np.sum(array_res_buy) == 2 and len(res_du) >= 1:
            res = np.random.randint(0, len(res_du))
            array_res_buy[res_du[res]] = 1
            #   print('Mua nguyên liệu_II:', array_res_buy, '----', res_du, card_infor, nl_bo_ra)
            return return_res_to_board(
                env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
            )

        if np.sum(array_res_buy) == 1 and len(res_du) >= 2:
            res = np.random.choice(res_du, 2, replace=False)
            array_res_buy[res] = 1
            #   print('Mua nguyên liệu_III:', array_res_buy, '----', res_du, res, card_infor, nl_bo_ra)
            return return_res_to_board(
                env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
            )

    if len(np.where(env_state[:90] == -(p_id + 1))[0]) < 3 and (action < 12):  # Úp thẻ
        card_can_upside_down = np.where(env_state[:90] == 5)[0]
        if len(card_can_upside_down) > 0:
            #   print('Úp thẻ', card_infor, card_infor, nl_bo_ra, p_id, get_id_card(card_id))
            env_state[card_id] = -(p_id + 1)
            if b_stocks[5] > 0:
                if np.sum(cur_p[:6]) == 10:
                    for res in np.argsort(nl_bo_ra):
                        if cur_p[res] > 0:
                            #   print('Trả nguyên liệu', res)
                            cur_p[res] -= 1
                            b_stocks[res] += 1
                            break
                if np.sum(cur_p[:6]) <= 9:
                    cur_p[5] += 1
                    b_stocks[5] -= 1

            env_state[card_id] = -(p_id + 1)
            if card_id < 40:
                env_state, lv1 = get_remove_card_on_lv_and_add_new_card(
                    env_state, lv1, p_id, 161, 3, card_id
                )
            elif card_id >= 40 and card_id < 70:
                env_state, lv2 = get_remove_card_on_lv_and_add_new_card(
                    env_state, lv2, p_id, 162, 3, card_id
                )
            else:
                env_state, lv3 = get_remove_card_on_lv_and_add_new_card(
                    env_state, lv3, p_id, 163, 3, card_id
                )

            env_state[107 + 12 * p_id : 119 + 12 * p_id] = cur_p
            env_state[101:107] = b_stocks
            return env_state, lv1, lv2, lv3

    action_have_res_count = np.array([-99])
    soluong_nl_bo_ra = np.array([99])
    action_co_the_lay_the = np.array([-1])

    for action_ in all_actions:  # Chọn những thẻ có thể lấy trên bàn và đang úp
        if action_ < 12:
            id_action_ = int(action_)
            id_card_normal_ = get_id_card_normal_in_lv(lv1, lv2, lv3)
        else:
            id_action_ = int(action_) - 12
            id_card_normal_ = np.where(env_state[:90] == -(p_id + 1))[0]

        card_id_ = id_card_normal_[id_action_]
        card_infor_ = normal_cards_infor[card_id_]
        card_price_ = card_infor_[-5:]
        nl_bo_ra_ = (card_price_ > cur_p[6:11]) * (card_price_ - cur_p[6:11])
        card_need_ = cur_p[:5] + cur_p[6:11] - card_price_

        if (
            -np.sum(card_need_[np.where(card_need_ < 0)]) <= cur_p[5]
            or np.min(card_need_) >= 0
        ):  # (x*x>0)
            action_co_the_lay_the = np.append(action_co_the_lay_the, action_)
            soluong_nl_bo_ra = np.append(soluong_nl_bo_ra, np.sum(nl_bo_ra_))
            #   print('Thẻ có thể lấy', get_id_card(card_id_), 'Số lượng nguyên liệu bỏ ra', np.sum(nl_bo_ra_))
            if (nl_bo_ra[card_infor_[1]] - cur_p[card_infor_[1]]) > 0:
                #   print('Thẻ này có nguyên liệu cần')
                action_have_res_count = np.append(action_have_res_count, action_)
            else:
                action_have_res_count = np.append(action_have_res_count, -99)

    action_chon_lay_the_co_nl_mac_dinh = np.where(action_have_res_count != -99)[0]
    check = False
    if len(action_chon_lay_the_co_nl_mac_dinh) > 0:  # Lấy thẻ có nguyên liệu mặc định
        nl_bo_ra_min = 99
        #   print("Lấy thẻ có nguyên liệu mặc định", soluong_nl_bo_ra, action_chon_lay_the_co_nl_mac_dinh)
        for action_id in action_chon_lay_the_co_nl_mac_dinh:
            if soluong_nl_bo_ra[action_id] < nl_bo_ra_min:
                nl_bo_ra_min = soluong_nl_bo_ra[action_id]
                action_ = action_have_res_count[action_id]
                check = True
        if check == True:
            return get_card(env_state, action_, cur_p, b_stocks, p_id, lv1, lv2, lv3)

    sum_res_need = 10 - np.sum(cur_p[:6])
    if sum_res_need >= 3:
        sum_res_need = 3

    res_board_have = b_stocks[:5]
    res_can_buy = np.where(((nl_bo_ra - cur_p[:5]) > 0) & (res_board_have > 0), 1, 0)
    array_res_buy = np.full(5, 0)

    for id, res in enumerate(res_can_buy):
        if res == 1:
            if np.sum(array_res_buy) == sum_res_need:
                break
            array_res_buy[id] = 1

    if np.sum(array_res_buy) > 1:
        #   print('Mua nguyên liệu_IIII:', array_res_buy, '----', res_can_buy, card_infor, nl_bo_ra)
        return return_res_to_board(
            env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
        )

    if len(soluong_nl_bo_ra) > 1 and np.min(soluong_nl_bo_ra) == 0:
        action_ = action_co_the_lay_the[np.argmin(soluong_nl_bo_ra)]
        #   print("Lấy thẻ không có nguyên liệu mặc định miễn phí:", np.min(soluong_nl_bo_ra))
        return get_card(env_state, action_, cur_p, b_stocks, p_id, lv1, lv2, lv3)

    if np.max(nl_bo_ra) > 0 and b_stocks[res_max] >= 4 and np.sum(cur_p[:6]) <= 8:
        #   print('Lấy 2 nguyên liệu_II:', res_max, '----', card_infor, nl_bo_ra)
        return return_2_res_to_board(
            env_state, lv1, lv2, lv3, cur_p, b_stocks, res_max, p_id
        )

    if np.sum(array_res_buy) > 0:
        for id, res in enumerate(b_stocks[:5]):
            if res > 0:
                if np.sum(array_res_buy) == sum_res_need:
                    break
                array_res_buy[id] = 1
        #   print('Mua nguyên liệu_IIIII:', array_res_buy, '----', res_can_buy, card_infor, nl_bo_ra)
        return return_res_to_board(
            env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
        )

    res_max = np.argmax(b_stocks[:5])
    if b_stocks[res_max] >= 4 and np.sum(cur_p[:6]) <= 8:
        #   print('Lấy 2 nguyên liệu_III, bất kỳ:', res_max, '----', card_infor, nl_bo_ra)
        return return_2_res_to_board(
            env_state, lv1, lv2, lv3, cur_p, b_stocks, res_max, p_id
        )

    for id, res in enumerate(b_stocks[:5]):
        if res > 0:
            if np.sum(array_res_buy) == sum_res_need:
                break
            array_res_buy[id] = 1

    if np.sum(array_res_buy) > 0:
        #   print('Mua nguyên liệu còn thừa trên bàn_IIIIII:', array_res_buy, '----', res_can_buy, card_infor, nl_bo_ra, sum_res_need)
        return return_res_to_board(
            env_state, lv1, lv2, lv3, cur_p, b_stocks, array_res_buy, p_id
        )

    if len(soluong_nl_bo_ra) > 1:
        action_ = action_co_the_lay_the[np.argmin(soluong_nl_bo_ra)]
        #   print("Lấy thẻ không có nguyên liệu mặc định:", np.min(soluong_nl_bo_ra))
        return get_card(env_state, action_, cur_p, b_stocks, p_id, lv1, lv2, lv3)

    #   print('Không làm gì cả')
    return env_state, lv1, lv2, lv3


@njit()
def getActionSize():
    return P_ACTION_SIZE


@njit()
def getAgentSize():
    return 4


@njit()
def getStateSize():
    return P_STATE_SIZE


@njit()
def getValidActions(player_state_origin: np.int64):
    list_action_return = np.zeros(15)
    list_action_return[: int(player_state_origin[P_COUNT_CAN_UPSIDE_DOWN])] = 1
    for id_card in range(3):
        card = player_state_origin[P_START_CARD_UPSIDE_DOWN:P_END_CARD_UPSIDE_DOWN][
            11 * id_card : 11 + 11 * id_card
        ]
        if np.sum(card) > 0:
            list_action_return[12 + id_card] = 1
    return list_action_return.astype(np.float64)


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env, lv1, lv2, lv3 = initEnv()
    _cc = 0

    while env[100] <= 400 and _cc <= 10000:
        idx = env[100] % 4
        player_state = getAgentState(env, lv1, lv2, lv3)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)

        list_action = getValidActions(player_state)
        if list_action[action] != 1:
            raise Exception("Action không hợp lệ")

        env, lv1, lv2, lv3 = stepEnv(action, env, lv1, lv2, lv3, list_action)
        if checkEnded(env) != 0:
            break

        _cc += 1

    for p_idx in range(4):
        env[100] = p_idx
        p_state = getAgentState(env, lv1, lv2, lv3)
        p_state[P_CLOSE_GAME] = 1
        if list_other[p_idx] == -1:
            act, per_player = p0(p_state, per_player)
            if getReward(p_state) == 1:
                winner = True
            else:
                winner = False
        elif list_other[p_idx] == 1:
            action, per1 = p1(p_state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(p_state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(p_state, per3)
        else:
            raise Exception("Sai list_other.")

    return winner, per_player


@njit()
def n_games_numba(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env, lv1, lv2, lv3 = initEnv()
    _cc = 0

    while env[100] <= 400 and _cc <= 10000:
        idx = env[100] % 4
        player_state = getAgentState(env, lv1, lv2, lv3)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)

        list_action = getValidActions(player_state)
        if list_action[action] != 1:
            raise Exception("Action không hợp lệ")

        env, lv1, lv2, lv3 = stepEnv(action, env, lv1, lv2, lv3, list_action)
        if checkEnded(env) != 0:
            break

        _cc += 1

    for p_idx in range(4):
        env[100] = p_idx
        p_state = getAgentState(env, lv1, lv2, lv3)
        p_state[P_CLOSE_GAME] = 1
        if list_other[p_idx] == -1:
            act, per_player = p0(p_state, per_player)
            if getReward(p_state) == 1:
                winner = True
            else:
                winner = False
        elif list_other[p_idx] == 1:
            action, per1 = p1(p_state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(p_state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(p_state, per3)
        else:
            raise Exception("Sai list_other.")

    return winner, per_player


def n_games_normal(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


@njit()
def check_run_under_njit(agent, perData):
    return True


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


def DataAgent():
    if env.getActionSize() < 150:
        perx_Chain3 = List(
            [
                np.zeros((env.getActionSize() ** 2, env.getActionSize())),
                np.argsort(
                    np.argsort(
                        np.random.rand(env.getActionSize() ** 2, env.getActionSize()),
                        axis=1,
                    ),
                    axis=1,
                )
                * 1.0
                + 1.0,
                np.zeros((1, 2)) - 1.0,
                np.zeros((1, 1)),
            ]
        )
    else:
        perx_Chain3 = List(
            [
                np.zeros((150**2, 150)),
                np.argsort(np.argsort(np.random.rand(150**2, 150), axis=1), axis=1)
                * 1.0
                + 1.0,
                np.zeros((1, 2)) - 1.0,
                np.zeros((1, 1)),
                np.zeros((1, env.getActionSize())),
                np.zeros((1, 150)),
                np.full((150**2, 150), np.arange(1, 151)).astype(np.float64),
            ]
        )
    return perx_Chain3


@njit
def argSortSpecial(w):
    indexTable = np.empty_like(w)
    indexTable2 = np.empty_like(w)
    for j in range(indexTable.shape[0]):
        indexTable[j, :] = np.argsort(w[j, :])
    for k in range(indexTable.shape[0]):
        indexTable2[k, :] = np.argsort(indexTable[k, :])
    return indexTable2 * 1.0


@njit
def generateArr(w):
    for i in range(w.shape[0]):
        np.random.shuffle(w[i, :])
    return w


@njit
def Train(state, per):
    #  if env.getReward(state)!=-1:
    #      if per[3][0][0]%100==0:
    #          print(per[3][0][0])
    if env.getActionSize() < 150:
        if per[2][0][0] == -1 or per[2][0][1] == -1:
            action = np.random.choice(np.where(env.getValidActions(state) == 1)[0])
            if per[2][0][0] == -1:
                per[2][0][0] = action
            elif per[2][0][1] == -1:
                per[2][0][1] = action
        else:
            weight = per[1][int(per[2][0][0] * env.getActionSize() + per[2][0][1])]
            action = np.argmax(weight * env.getValidActions(state))
            per[2][0][0] = per[2][0][1]
            per[2][0][1] = action
        if env.getReward(state) != -1:
            per[3][0][0] += 1
            if env.getReward(state) == 1:
                per[0] += per[1] * 1.0
            else:
                per[1] = (
                    argSortSpecial(
                        np.random.rand(env.getActionSize() ** 2, env.getActionSize())
                    )
                    * 1.0
                    + 1.0
                )

        return int(action), per
    else:
        if per[3][0][0] < 10000:
            action = np.random.choice(np.where(env.getValidActions(state) == 1)[0])
            per[4][0][int(action)] += 1
            if env.getReward(state) != -1:
                per[3][0][0] += 1
                if per[3][0][0] == 10000:
                    per[5][0] = np.arange(env.getActionSize())[
                        np.argsort(per[4][0, :])
                    ][-150:]
        else:
            if per[2][0][0] == -1 or per[2][0][1] == -1:
                action = np.random.choice(np.where(env.getValidActions(state) == 1)[0])
                if per[2][0][0] == -1:
                    per[2][0][0] = action
                elif per[2][0][1] == -1:
                    per[2][0][1] = action
            else:
                if (
                    np.where(per[5][0] == per[2][0][0])[0].shape[0] == 0
                    or np.where(per[5][0] == per[2][0][1])[0].shape[0] == 0
                ):
                    action = np.random.choice(
                        np.where(env.getValidActions(state) == 1)[0]
                    )
                    per[2][0][0] = per[2][0][1]
                    per[2][0][1] = action
                else:
                    id1 = np.where(per[5][0] == per[2][0][0])[0][0]
                    id2 = np.where(per[5][0] == per[2][0][1])[0][0]
                    idx = int(id1 * env.getActionSize() + id2)
                    if idx >= 0 and idx <= 150**2:
                        weight = per[1][int(id1 * 150 + id2)]
                        list_action = (
                            weight
                            * env.getValidActions(state)[per[5][0].astype(np.int64)]
                        )
                        action = per[5][0][np.argmax(list_action)]
                        if env.getValidActions(state)[int(action)] == 0:
                            action = np.random.choice(
                                np.where(env.getValidActions(state) == 1)[0]
                            )
                        per[2][0][0] = per[2][0][1]
                        per[2][0][1] = action
                    else:
                        action = np.random.choice(
                            np.where(env.getValidActions(state) == 1)[0]
                        )
                        per[2][0][0] = per[2][0][1]
                        per[2][0][1] = action
            if env.getReward(state) != -1:
                per[3][0][0] += 1
                if env.getReward(state) == 1:
                    per[0] += per[1] * 1.0
                else:
                    per[1] = generateArr(per[6])
        return int(action), per


@njit
def Test(state, per):
    validActions = env.getValidActions(state)
    actions = np.where(validActions == 1)[0]
    actionSize = env.getActionSize()
    if actionSize < 150:
        if per[1][0][0] == -1 or per[1][0][1] == -1:
            action = actions[np.random.randint(0, actions.shape[0])]
            if per[1][0][0] == -1:
                per[1][0][0] = action
            elif per[1][0][1] == -1:
                per[1][0][1] = action
        else:
            weight = per[0][int(per[1][0][0] * actionSize + per[1][0][1])]
            action = np.argmax(weight * validActions + validActions)
            per[1][0][0] = per[1][0][1]
            per[1][0][1] = action
        return action, per
    else:
        if per[1][0][0] == -1 or per[1][0][1] == -1:
            action = actions[np.random.randint(0, actions.shape[0])]
            if per[1][0][0] == -1:
                per[1][0][0] = action
            elif per[1][0][1] == -1:
                per[1][0][1] = action
        else:
            if per[1][0][0] == -1 or per[1][0][1] == -1:
                action = actions[np.random.randint(0, actions.shape[0])]
                if per[1][0][0] == -1:
                    per[1][0][0] = action
                elif per[1][0][1] == -1:
                    per[1][0][1] = action
            else:
                id1_ = np.where(per[2][0] == per[1][0][0])[0]
                id2_ = np.where(per[2][0] == per[1][0][1])[0]
                if id1_.shape[0] == 0 or id2_.shape[0] == 0:
                    action = actions[np.random.randint(0, actions.shape[0])]
                    per[1][0][0] = per[1][0][1]
                    per[1][0][1] = action
                else:
                    id1 = id1_[0]
                    id2 = id2_[0]
                    idx = id1 * actionSize + id2
                    if idx >= 0 and idx <= 150**2:
                        weight = per[0][int(id1 * 150 + id2)]
                        list_action = weight * validActions[per[2][0].astype(np.int64)]
                        action = per[2][0][np.argmax(list_action)]
                        if validActions[int(action)] == 0:
                            action = actions[np.random.randint(0, actions.shape[0])]
                        per[1][0][0] = per[1][0][1]
                        per[1][0][1] = action
                    else:
                        action = actions[np.random.randint(0, actions.shape[0])]
                        per[1][0][0] = per[1][0][1]
                        per[1][0][1] = action
        return int(action), per


def convert_to_save(perData):
    data = List()
    if env.getActionSize() < 150:
        if len(perData) == 2:
            raise Exception("Data này đã được convert rồi.")
        data.append(perData[0])
        data.append(perData[2])
    else:
        if len(perData) == 3:
            raise Exception("Data này đã được convert rồi.")
        data.append(perData[0])
        data.append(perData[2])
        data.append(perData[5])

    return data


def convert_to_test(perData):
    return List(perData)
